# This file is part of pymadcad,  distributed under license LGPL v3

'''  This modules provide tools for accessing data using its space location (not for final user).
	
	The complexity and therefore the cost of those operations are most of the time close to the hashmap complexity O(1). It means data is found in time independently of the actual size of the mesh or whatever storage it is.
'''

from .mathutils import *
from . import core
from . import mesh
from functools import reduce
from math import floor, ceil, sqrt, inf


# ------ connectivity tools -------
		

def edgekey(a,b):
	''' Return a key for a non-directional edge '''
	if a < b:	return (a,b)
	else:		return (b,a)
	
def facekeyo(a,b,c):
	''' Return a key for an oriented face '''
	if a < b and b < c:		return (a,b,c)
	elif a < b:				return (c,a,b)
	else:					return (b,c,a)
	
def arrangeface(f, p):
	''' Return the face indices rotated the way `p` is the first index, if `p` is in the face '''
	if   p == f[1]:	return f[1],f[2],f[0]
	elif p == f[2]:	return f[2],f[0],f[1]
	else:			return f
	
def arrangeedge(e, p):
	''' Return the edge indices rotated the way `p` is the first index, if `p` is in the edge '''
	if p == e[1]:	return e[1], e[0]
	else:			return e

def connpp(ngons):
	''' Point to point connectivity 
		
		input is a list of ngons (tuple of 2 to n indices)
	'''
	conn = {}
	for loop in ngons:
		for i in range(len(loop)):
			for a,b in ((loop[i-1],loop[i]), (loop[i],loop[i-1])):
				if a not in conn:		conn[a] = [b]
				elif b not in conn[a]:	conn[a].append(b)
	return conn
	
def connef(faces):
	''' Oriented edge to face connectivity '''
	conn = {}
	for i,f in enumerate(faces):
		for e in ((f[0],f[1]), (f[1],f[2]), (f[2],f[0])):
			conn[e] = i
	return conn
	
def connpe(edges):
	''' Point to edge connectivity '''
	conn = Asso()
	for i,edge in enumerate(edges):
		for p in edge:
			conn.add(p,i)
	return conn


def connexity(links):
	''' Return the number of links referencing each point as a dictionary {point: num links} '''
	reach = {}
	for l in links:
		for p in l:
			reach[p] = reach.get(p,0) +1
	return reach

def suites(lines, oriented=True, cut=True, loop=False):
	''' Return a list of the suites that can be formed with lines.
		`lines` is an iterable of edges
		
		Parameters:
			oriented:      specifies that (a,b) and (c,b) will not be assembled
			cut:           cut suites when they are crossing each others
		
		Return a list of the sequences that can be formed
	'''
	# TODO: rewrite this to use hashmaps instead of search loops
	lines = list(lines)
	# get contiguous suite of points
	suites = []
	while lines:
		suite = list(lines.pop())
		found = True
		while found:
			found = False
			for i,edge in enumerate(lines):
				if edge[-1] == suite[0]:		suite[0:1] = edge
				elif edge[0] == suite[-1]:		suite[-1:] = edge
				# for unoriented lines
				elif not oriented and edge[0] == suite[0]:		suite[0:1] = reversed(edge)
				elif not oriented and edge[-1] == suite[-1]:	suite[-1:] = reversed(edge)
				else:
					continue
				lines.pop(i)
				found = True
				break
			if loop and suite[-1] == suite[0]:	break
		suites.append(suite)
	# cut at suite intersections (sub suites or crossing suites)
	if cut:
		reach = {}
		for suite in suites:
			for p in suite:
				reach[p] = reach.get(p,0) + 1
		for suite in suites:
			for i in range(1,len(suite)-1):
				if reach[suite[i]] > 1:
					suites.append(suite[i:])
					suite[i+1:] = []
					break
	return suites
	



# --- hashed storages  -----

class PositionMap:
	''' Holds objects associated with their location.
		Every object can be bound to multiple locations, and each location can hold multiple objects.
		`cellsize` defines the box size for location hashing (the smaller it is, the bigger the memory footprint will be for non-point primitives)
		
		Attributes defined here:
			:cellsize:    the boxing parameter (DON'T CHANGE IT IF NON-EMPTY)
			:dict:        the hashmap from box to objects lists
	'''
	__slots__ = 'cellsize', 'dict', 'options'
	
	def __init__(self, cellsize, iterable=None):
		self.options = {}
		self.cellsize = cellsize
		self.dict = {}
		if iterable:	self.update(iterable)

	def keysfor(self, space):
		''' Rasterize the primitive, yielding the successive position keys 
			currently allowed primitives are 

				:points:		vec3
				:segments:	(vec3,vec3)
				:triangles:	(vec3,vec3,vec3)
		'''
		cell = self.cellsize
		# point
		if isinstance(space, vec3):
			yield tuple(i64vec3(glm.floor(space/cell)))
		
		# segment
		elif isinstance(space, tuple) and len(space) == 2:
			# permutation of coordinates to get the direction the closer to Z
			n = glm.abs(space[1]-space[0])
			if max(n) < NUMPREC*max((max(p) for p in space)):	return
			if   n[1] >= n[0] and n[1] >= n[2]:	order,reorder = (2,0,1),(1,2,0)
			elif n[0] >= n[1] and n[0] >= n[2]:	order,reorder = (1,2,0),(2,0,1)
			else:								order,reorder = (0,1,2),(0,1,2)
			space = [ vec3(p[order[0]], p[order[1]], p[order[2]])	for p in space]
			
			# prepare variables
			cell2 = cell/2
			v = space[1]-space[0]
			dy = v[1]/v[2]
			dx = v[0]/v[2]
			o = space[0]
			fy = lambda z:	o[1] + dy*(z-o[2])
			fx = lambda z:	o[0] + dx*(z-o[2])
			
			# z selection
			if v[2] > 0:	zmin,zmax = space[0][2], space[1][2]
			else:			zmin,zmax = space[1][2], space[0][2]
			zmin -= zmin%cell
			for i in range(max(1,ceil((zmax-zmin)/cell))):
				z = zmin + cell*i + cell2
				
				# y selection
				if dy > 0:	ymin,ymax = fy(z-cell2), fy(z+cell2)
				else:		ymin,ymax = fy(z+cell2), fy(z-cell2)
				ymin -= ymin%cell
				for j in range(max(1,ceil((ymax-ymin)/cell))):
					y = ymin + j*cell + cell2
					
					# x selection
					if dx > 0:	xmin,xmax = fx(z-cell2), fx(z+cell2)
					else:		xmin,xmax = fx(z+cell2), fx(z-cell2)
					xmin -= xmin%cell
					for k in range(max(1,ceil((xmax-xmin)/cell))):
						x = xmin + k*cell + cell2
						
						p = (x,y,z)
						yield tuple([floor(p[i]/cell)		for i in reorder])
		
		# triangle
		elif isinstance(space, tuple) and len(space) == 3:
			# permutation of coordinates to get the normal the closer to Z
			n = glm.abs(cross(space[1]-space[0], space[2]-space[0]))
			if max(n) < NUMPREC*max((max(p) for p in space)):	return
			if   n[1] >= n[0] and n[1] >= n[2]:	order,reorder = (2,0,1),(1,2,0)
			elif n[0] >= n[1] and n[0] >= n[2]:	order,reorder = (1,2,0),(2,0,1)
			else:								order,reorder = (0,1,2),(0,1,2)
			space = [ vec3(p[order[0]], p[order[1]], p[order[2]])	for p in space]
			
			# prepare variables
			v = [space[i-1]-space[i]	for i in range(3)]
			n = cross(v[0],v[1])
			dx = -n[0]/n[2]
			dy = -n[1]/n[2]
			o = space[0]
			fz = lambda x,y:	o[2] + dx*(x-o[0]) + dy*(y-o[1])
			cell2 = cell/2
			pmin = reduce(glm.min, space)
			pmax = reduce(glm.max, space)
			xmin,xmax = pmin[0],pmax[0]
			pmin -= pmin%cell
			pmax += cell - pmax%cell
			
			# x selection
			xmin -= xmin%cell
			for i in range(max(1,ceil((xmax-xmin)/cell))):
				x = xmin + cell*i + cell2
			
				# y selection
				cand = []
				for i in range(3):
					# NOTE: cet interval ajoute parfois des cases inutiles apres les sommets
					if (space[i-1][0]-x+cell2)*(space[i][0]-x-cell2) <= 0 or (space[i-1][0]-x-cell2)*(space[i][0]-x+cell2) <= 0:
						d = v[i][1]/(v[i][0] if v[i][0] else inf)
						cand.append( space[i][1] + d * (x-cell2-space[i][0]) )
						cand.append( space[i][1] + d * (x+cell2-space[i][0]) )
				ymin,ymax = max(pmin[1],min(cand)), min(pmax[1],max(cand))
				ymin -= ymin%cell
				for j in range(max(1,ceil((ymax-ymin)/cell))):
					y = ymin + cell*j + cell2
				
					# z selection
					cand = []
					cand.append( fz(x-cell2, y-cell2) )
					cand.append( fz(x+cell2, y-cell2) )
					cand.append( fz(x-cell2, y+cell2) )
					cand.append( fz(x+cell2, y+cell2) )
					zmin,zmax = max(pmin[2],min(cand)), min(pmax[2],max(cand))
					zmin -= zmin%cell
					for k in range(max(1,ceil((zmax-zmin)/cell))):
						z = zmin + cell*k + cell2
						
						# remove box from corners that goes out of the area
						p = (x,y,z)
						if pmin[0]<p[0] and pmin[1]<p[1] and pmin[2]<p[2] and p[0]<pmax[0] and p[1]<pmax[1] and p[2]<pmax[2]:
							yield tuple([floor(p[i]/cell)	for i in reorder])
		
		else:
			raise TypeError("PositionMap only supports keys of type:  points, segments, triangles")
	
	def keysfor(self, space):
		''' Rasterize the primitive, yielding the successive position keys 
			currently allowed primitives are 

				:points:		vec3
				:segments:		(vec3,vec3)
				:triangles:		(vec3,vec3,vec3)
		'''
		# point
		if isinstance(space, vec3):
			return tuple(i64vec3(glm.floor(space/cell))),
		# segment
		elif isinstance(space, tuple) and len(space) == 2:
			return core.rasterize_segment(space, self.cellsize)
		# triangle
		elif isinstance(space, tuple) and len(space) == 3:
			return core.rasterize_triangle(space, self.cellsize)
		else:
			raise TypeError("PositionMap only supports keys of type:  points, segments, triangles")
	
	def update(self, other):
		''' Add the elements from an other PositionMap or from an iterable '''
		if isinstance(other, PositionMap):
			assert self.cellsize == other.cellsize
			for k,v in other.dict.items():
				if k in self.dict:	self.dict[k].extend(v)
				else:				self.dict[k] = v
		elif hasattr(other, '__iter__'):
			for space,obj in other:
				self.add(space,obj)
		else:
			raise TypeError("update requires a PositionMap or an iterable of couples (space, obj)")
	
	def add(self, space, obj):
		''' 
			add an object associated with a primitive 
			check `keysfor` for a description of allowed primitives
		'''
		for k in self.keysfor(space):
			if k not in self.dict:	self.dict[k] = [obj]
			else:					self.dict[k].append(obj)
	
	def get(self, space):
		''' 
			get the objects potentially intersecting the given primitive 
			check `keysfor` for a description of allowed primitives
		'''
		for k in self.keysfor(space):
			if k in self.dict:
				yield from self.dict[k]
				
	def __contains__(self, space):
		''' check if any stored object is potentially intersecting the given primitive '''
		return next(self.get(space), None) is not None
	
	_display = (
		[	
			vec3(0,0,0),
			vec3(1,0,0),vec3(0,1,0),vec3(0,0,1),
			vec3(0,1,1),vec3(1,0,1),vec3(1,1,0),
			vec3(1,1,1)],
		[	
			(0,1),(0,2),(0,3),
			(1,6),(2,6),
			(1,5),(3,5),
			(2,4),(3,4),
			(4,7),(5,7),(6,7),
			],
		)
	def display(self, scene):
		web = mesh.Web()
		if 'color' in self.options:		web.options['color'] = self.options['color']
		base = vec3(self.cellsize)
		for k in self.dict:
			l = len(web.points)
			web += mesh.Web([base*(p+k)  for p in self._display[0]], self._display[1], groups=[k])
		return web.display(scene)

def meshcellsize(mesh):
	''' Returns a good cell size to index primitives of a mesh with a PositionMap 
		
		See implementation.
	'''
	# the number of key for triangles is propotionate to the surface
	# points are placed on this surface, so sqrt(len(pts)) is a good hint for the point density on the surface
	return length(mesh.box().width) / sqrt(len(mesh.points))


class PointSet:
	''' Holds a list of points and hash them.
		The points are holds using indices, that allows to get the point buffer at any time, or to retrieve only a point index.
		`cellsize` defines the box size in which two points are considered to be the same
		
		Methods are inspired from the builtin type set
		
		Attributes defined here:
			:points:     the point buffer (READ-ONLY PURPOSE)
			:cellsize:   the boxing parameter (DON'T CHANGE IT IF NON-EMPTY). mendatory and is the distance at which you want to distinguish points
			:dict:       the hashmap from box to point indices
			
		Build parameters:
			:iterable:    use it to build the set by inserting elements
			:manage:      pass a list for inplace use it, only indexing will be built
	'''
	__slots__ = 'points', 'cellsize', 'dict'
	
	def __init__(self, cellsize, iterable=None, manage=None):
		self.cellsize = cellsize
		self.dict = {}
		if manage is not None:
			assert iterable is None
			self.points = manage
			for i in reversed(range(len(self.points))):
				self.dict[self.keyfor(self.points[i])] = i
		else:
			self.points = typedlist(dtype=vec3)
			if iterable:	self.update(iterable)
	
	def keyfor(self, pt):
		''' 
			hash key for a point. 
			points with the same key will be considered equivalent and merged in the set 
		'''
		return tuple(i64vec3(glm.floor(pt/self.cellsize)))
		
	def keysfor(self, pt):
		''' iterable of positions at which an equivalent to the argument point can be '''
		vox = pt/self.cellsize
		k = i64vec3(glm.floor(vox-0.5+NUMPREC)), i64vec3(glm.floor(vox+0.5-NUMPREC))
		return (
			(k[0][0], k[0][1], k[0][2]),
			(k[1][0], k[0][1], k[0][2]),
			(k[0][0], k[1][1], k[0][2]),
			(k[1][0], k[1][1], k[0][2]),
			(k[0][0], k[0][1], k[1][2]),
			(k[1][0], k[0][1], k[1][2]),
			(k[0][0], k[1][1], k[1][2]),
			(k[1][0], k[1][1], k[1][2]),
			)
	
	def update(self, iterable):
		''' add points from an iterable '''
		for pt in iterable:	self.add(pt)
	def difference_update(self, iterable):
		''' Remove the points from an iterable '''
		for pt in iterable:	self.discard(pt)
		
	def add(self, pt) -> int:
		''' 
			add a point to the set if no equivalent point is at this position 
			return the index of the created point or the pre-existing point at this position.
		'''
		for key in self.keysfor(pt):
			if key in self.dict:
				return self.dict[key]
		self.dict[self.keyfor(pt)] = l = len(self.points)
		self.points.append(pt)
		return l
	def remove(self, pt) -> int:
		''' 
			remove the point at given position point from the set, returning its former index.
			raise if no point exist at this position
		'''
		for key in self.keysfor(pt):
			if key in self.dict:
				return self.dict.pop(key)
		else:					
			raise IndexError("position doesn't exist in set")
	def discard(self, pt):
		''' remove all points at positions equivalent to the given location (if any) '''
		for key in self.keysfor(pt):
			if key in self.dict:
				del self.dict[key]
	
	def __contains__(self, pt) -> bool:
		''' true if there is a point at the given location in the set '''
		for key in self.keysfor(pt):
			if key in self.dict:	return True
		return False
	def __getitem__(self, pt) -> int:
		''' return the index of the point at given location in the set '''
		for key in self.keysfor(pt):
			if key in self.dict:	return self.dict[key]
		raise IndexError("position doesn't exist in set")
		
	__iadd__ = update
	__isub__ = difference_update
	def __add__(self, iterable):
		s = PositionSet()
		s.union(self.points)
		s.union(iterable)
		return s
	def __sub__(self, iterable):
		s = PositionSet()
		s.union(self.points)
		s.difference(iterable)
		return s


class Asso(object):
	''' Associative container.
		This is a sort dict that stores as many values as we want for each key. The return value for a key is then always an iterable of the associated values, even when the value set to the key is unique and is not an iterator.
		
		Asso(iterable)				the iterable must yields (key,value) pairs
		Asso(keys=[], values=[])	build form existing lists, assuming they are already sorted
	'''
	__slots__ = '_table'
	
	def __init__(self, iter=None):
		if isinstance(iter, Asso):
			self._table = dict(iter.table)
		else:
			self._table = {}
			if iter:
				self.update(iter)
		
	def __getitem__(self, key):
		''' return an iterable of the objects associated to the given key '''
		i = 0
		while True:
			k = (i,key)
			if k not in self._table:
				break
			yield self._table[k]
			i += 1
			
	def add(self, key, value):
		''' associate a new value to this key '''
		i = 0
		while True:
			k = (i,key)
			if k not in self._table:
				self._table[k] = value
				return
			i += 1
	
	def remove(self, key, value):
		l = self.connexity(key)-1
		i = l
		while i>=0:
			k = (i,key)
			if self._table[k] == value:
				last = (l,key)
				self._table[k] = self._table[last]
				del self._table[last]
				return
			i -= 1
		raise KeyError('key not in table')
		
	def discard(self, key, value):
		''' remove all (key,value) pair of that table '''
		l = self.connexity(key)-1
		i = l
		while i>=0:
			k = (i,key)
			if self._table[k] == value:
				last = (l,key)
				self._table[k] = self._table[last]
				del self._table[last]
				l -= 1
			i -= 1
	
	def update(self, other):
		''' append all key,value associations to this Asso '''
		if isinstance(other, dict):
			other = other.items()
		for k,v in other:
			self.add(k,v)
		
	def __add__(self, other):
		new = Asso()
		new._table = dict(self._table)
		new.update(other)
		return new
		
		
	def __delitem__(self, key):
		''' delete all associations with this key '''
		i = 0
		while True:
			k = (i,k)
			if k not in self._table:
				break
			del self._table[k]
			i += 1
	
	def clear(self):
		''' empty the container '''
		self._table.clear()

	def items(self):
		''' iterator of (key, value) pairs '''
		for k,v in self._table.items():
			yield k[1],v

	def keys(self):
		''' iterator of the keys '''
		for k,v in self._table:
			yield k[1]
		
	def values(self):
		''' iterator of the values '''
		return self._table.values()
		
	def connexity(self, key):
		''' return the number of values associated to the given key '''
		i = 0
		while True:
			k = (i,key)
			if k not in self._table:
				break
			i += 1
		return i
		

	def __contains__(self, key):
		''' return True if key is associated to something '''
		return (0,key) in self._table
		
	def __repr__(self):
		return 'Asso([{}])'.format(', '.join(
					'({}, {})'.format(repr(k), repr(v))	
						for k,v in self.items()
					))

def test_asso():
	m = Asso([
			(1,'truc'), 
			(2,'machin'),
			(12, 'chose'),
			(1, 'bidule'),
			])
	nprint('created', m)
	
	m.update([
		(1,'machin'),
		(-1,'gna'),
		])
	nprint('updated', m)
	
	m.add(1, 2)
	nprint('inserted', m)
	
	assert set(m[1]) == {'truc', 'bidule', 2, 'machin'}
	assert set(m[12]) == {'chose'}





# This file is part of pymadcad,  distributed under license LGPL v3
__all__ = ['Solid', 'placement', 'explode', 'explode_offsets']

from ..mathutils import *


class Solid:
	''' Solid for objects display
	
		A Solid is also a way to group objects and move it anywhere without modifying them, as the objects contained in a solid are considered to be in solid local coordinates.
		A Solid is just like a dictionary with a pose.
	
		Attributes:
			pose (mat4):   placement matrix, it defines a base in which the solid's content is displayed
			content (dict/list):      objects to display using the solid's pose
			
		Example:
			
			>>> mypart = icosphere(vec3(0), 1)
			>>> s = Solid(part=mypart, anything=vec3(0))   # create a solid with whatever inside
			
			>>> s.transform(vec3(1,2,3))   # make a new translated solid, keeping the same content without copy
			
			>>> # put any content in as a dict
			>>> s['part']
			<Mesh ...>
			>>> s['whatever'] = vec3(5,2,1)
	'''
	
	def __init__(self, pose=1, **content):
		self.pose = mat4(pose)
		self.content = content
		
	def __copy__(self):
		return Solid(self.pose, **self.content)
	
	def transform(self, value) -> 'Solid':
		''' Displace the solid by the transformation '''
		return Solid(transform(value) * self.pose, **self.content)
	
	def place(self, *args, **kwargs) -> 'Solid': 
		''' Strictly equivalent to `.transform(placement(...))`, see `placement` for parameters specifications. '''
		return Solid(placement(*args, **kwargs), **self.content)
		
	def loc(self, *args):
		transform = mat4()
		for key in args:
			transform = transform @ obj.pose
		return transform
	
	def deloc(self, *args):
		obj = self
		transform = mat4()
		for key in args:
			transform = transform @ obj.pose
			obj = obj.content[key]
		return obj.transform(transform)
	
	def reloc(self, *args):
		indev
	
	# convenient content access
	def __getitem__(self, key):
		''' Shorthand to `self.content` '''
		return self.content[key]
		
	def __setitem__(self, key, value):
		''' Shorthand to `self.content` '''
		self.content[key] = value
	
	def append(self, value):
		''' Add an item in self.content, a key is automatically created for it and is returned '''
		key = next(i 	for i in range(len(self.content)+1)
						if i not in self.content	)
		self.content[key] = value
		return key
	
	def set(self, **objs):
		''' Convenient method to set many elements in one call.
			Equivalent to `self.content.update(objs)`
		'''
		self.content.update(objs)
		return self
	
	def display(self, scene):
		from .displays import SolidDisplay
		return SolidDisplay(scene, self)


def placement(*pairs, precision=1e-3):
	''' Return a transformation matrix that solved the placement constraints given by the surface pairs
	
		Parameters:
		
			pairs:	a list of pairs to convert to kinematic joints
					
					- items can be couples of surfaces to convert to joints using `guessjoint`
					- tuples (joint_type, a, b)  to build joints `joint_type(solida, solidb, a, b)`
			
			precision: surface guessing and kinematic solving precision (distance)
		
		Each pair define a joint between the two assumed solids (a solid for the left members of the pairs, and a solid for the right members of the pairs). Placement will return the pose of the first relatively to the second, satisfying the constraints.
		
		Example:
		
			>>> # get the transformation for the pose
			>>> pose = placement(
			...		(screw['part'].group(0), other['part'].group(44)),  # two cylinder surfaces: Cylindrical joint
			...		(screw['part'].group(4), other['part'].group(25)),    # two planar surfaces: Planar joint
			...		)  # solve everything to get solid's pose
			>>> # apply the transformation to the solid
			>>> screw.pose = pose
			
			>>> # or
			>>> screw.place(
			...		(screw['part'].group(0), other['part'].group(44)),
			...		(screw['part'].group(4), other['part'].group(25)),
			...		)
			
			>>> screw.place(
			...		(Revolute, screw['axis'], other['screw_place']),
			...		)
	'''
	from ..reverse import guessjoint
	
	joints = []
	for pair in pairs:
		if len(pair) == 2:		joints.append(guessjoint((0, 1), *pair, precision*0.25))
		elif len(pair) == 3:	joints.append(pair[0]((0, 1), *pair[1:]))
		else:
			raise TypeError('incorrect pair definition', pair)
	
	if len(joints) > 1:
		kin = Kinematic(joints)
		parts = kin.parts(kin.solve())
		return affineInverse(parts[1]) * parts[0]
	else:
		return affineInverse(joints[0].direct(joints[0].default))
	
def convexhull(pts):
	import scipy.spatial
	if len(pts) == 3:
		return Mesh(pts, [(0,1,2),(0,2,1)])
	elif len(pts) > 3:
		hull = scipy.spatial.ConvexHull(typedlist_to_numpy(pts, 'f8'))
		m = Mesh(pts, hull.simplices.tolist())
		return m
	else:
		return Mesh(pts)

			
def extract_used(obj):
	if isinstance(obj, Mesh):	links = obj.faces
	elif isinstance(obj, Web):	links = obj.edges
	elif isinstance(obj, Wire):	links = [obj.indices]
	else:
		raise TypeError('obj must be a mesh of any kind')
	
	return striplist(obj.points[:], links)[0]

	
def explode_offsets(solids) -> '[(solid_index, parent_index, offset, barycenter)]':
	''' Build a graph of connected objects, ready to create an exploded view or any assembly animation.
		See `explode()` for an example. The exploded view is computed using the meshes contained in the given solids, so make sure there everything you want in their content.
	
		Complexity is `O(m * n)` where m = total number of points in all meshes, n = number of solids
		
		NOTE:
			
			Despite the hope that this function will be helpful, it's (for computational cost reasons) not a perfect algorithm for complex assemblies (the example above is at the limit of a simple one). The current algorithm will work fine for any simple enough assembly but may return unexpected results for more complex ones.
		
	'''
	import scipy.spatial.qhull
	# build convex hulls
	points = [[] for s in solids]
	# recursively search for meshes in solids
	def process(i, solid):
		if hasattr(solid.content, 'values'):	it = solid.content.values()
		else:									it = solid.content
		for obj in solid.content.values():
			if isinstance(obj, Solid):
				process(i, obj)
			elif isinstance(obj, (Mesh,Web,Wire)):
				try:
					points[i].extend(extract_used(convexhull(extract_used(obj)).transform(solid.pose)))
				except scipy.spatial.qhull.QhullError:
					continue
			
	for i,solid in enumerate(solids):
		process(i,solid)
	
	# create convex hulls and prepare for parenting
	hulls = [convexhull(pts).orient()  for pts in points]
	boxes = [hull.box()  for hull in hulls]
	normals = [hull.vertexnormals()  for hull in hulls]
	barycenters = [hull.barycenter()  for hull in hulls]
	
	scores = [inf] * len(solids)
	parents = [None] * len(solids)
	offsets = [vec3(0)] * len(solids)
	
	# build a graph of connected things (distance from center to convex hulls)
	for i in range(len(solids)):
		center = barycenters[i]	
		for j in range(len(solids)):
			if i == j:
				continue
			
			# case of non-connection, the link won't appear in the graph
			if boxes[i].intersection(boxes[j]).isempty():
				continue
			
			# the parent is always the biggest of the two, this also breaks any possible parenting cycle
			if length2(boxes[i].width) > length2(boxes[j].width):
				continue
			
			# select the shortest link
			d, prim = distance2_pm(center, hulls[j])
			
			if d < scores[i]:
				# take j as new parent for i
				scores[i] = d
				parents[i] = j
				
				# get the associated displacement vector
				pts = hulls[j].points
				if isinstance(prim, int):
					normal = normals[j][prim]
					offsets[i] = center - pts[prim]
				elif len(prim) == 2:
					normal = normals[j][prim[0]] + normals[j][prim[1]]
					offsets[i] = noproject(center - pts[prim[0]],  
										pts[prim[0]]-pts[prim[1]])
				elif len(prim) == 3:
					normal = cross(pts[prim[1]]-pts[prim[0]], pts[prim[2]]-pts[prim[0]])
					offsets[i] = project(center - pts[prim[0]],  normal)
				else:
					raise AssertionError('prim should be an index for  point, face, triangle')
				if dot(offsets[i], normal) < 0:
					offsets[i] = -offsets[i]
	
	# resolve dependencies to output the offsets in the resolution order
	order = []
	reached = [False] * len(solids)
	i = 0
	while i < len(solids):
		if not reached[i]:
			j = i
			chain = []
			while not (j is None or reached[j]):
				reached[j] = True
				chain.append(j)
				j = parents[j]
			order.extend(reversed(chain))
		i += 1
	
	# move more parents that have children on their way out				
	blob = [deepcopy(box) 	for box in boxes]
	for i in reversed(range(len(solids))):
		j = parents[i]
		if j and length2(offsets[i]):
			offsets[i] *= (1 
							+ 0.5* length(blob[i].width) / length(offsets[i]) 
							- dot(blob[i].center - barycenters[i], offsets[i]) / length2(offsets[i])
							)
			blob[j].union_update(blob[i].transform(offsets[i]))
								
	return [(i, parents[i], offsets[i], barycenters[i])  for i in order]
			
	
def explode(solids, factor=1, offsets=None) -> '(solids:list, graph:Mesh)':
	''' Move the given solids away from each other in the way of an exploded view.
		It makes easier to seen the details of an assembly . See `explode_offsets` for the algorithm.
		
		Parameters:
			
			solids:		a list of solids (copies of each will be made before displacing)
			factor:		displacement factor, 0 for no displacement, 1 for normal displacement
			offsets:	if given, must be the result of `explode_offsets(solids)`
		
		Example:
		
			>>> # pick some raw model and separate parts
			>>> imported = read(folder+'/some_assembly.stl')
			>>> imported.mergeclose()
			>>> parts = []
			>>> for part in imported.islands():
			...     part.strippoints()
			...     parts.append(Solid(part=segmentation(part)))
			... 
			>>> # explode the assembly to look into it
			>>> exploded = explode(parts)
		
	'''
	solids = [copy(solid)  for solid in solids]
	if not offsets:
		offsets = explode_offsets(solids)
	
	graph = Web(groups=[None])
	shifts = [	(solids[solid].position - solids[parent].position)
				if parent else vec3(0)
				for solid, parent, offset, center in offsets]
	for solid, parent, offset, center in offsets:
		if parent:
			solids[solid].position = solids[parent].position + shifts[solid] + offset * factor
			
			graph.edges.append((len(graph.points), len(graph.points)+1))
			graph.tracks.append(0)
			graph.points.append(solids[parent].position + shifts[solid] + center)
			graph.points.append(solids[solid].position + center)
			
	return [solids, graph]



# This file is part of pymadcad,  distributed under license LGPL v3

'''     pymadcad
		it's time to throw parametric softwares out !
		
Simple yet powerful CAD (Computer Aided Design) library, written with Python.

main concepts
-------------

Current CAD softwares, are based on a full GUI oriented approach, with parametric features bases on dynamic relationship between the scene objects. This can leads to many problems:
	
	- Objects relationships are never visible, thus tricky to understand and to manage. On most of these softwares it's in fact impossible to manage it.
	
	- The software has to be very complex to handle all possible behaviors in parameters change in order to update the model.
	
	- The GUI has to be very complex to allow the user to deal with every aspect of the objects (thus many tools, many object menus, sub-menus, context-dependent behaviors). The softwares are very hard to maintain, and often buggy
	
	- When a CAD file grow in complexity, we must avoid to modify it deeply or the entire build-tree fails and it's no more possible to repair.
	
	- The build tree is rarely human-readable.
	
The current library differs in its approach by the following points:

	- MADCAD is a script based CAD
		all the functions of madcad are available in the current API. Nothing is hidden.
		The python script is the build tree, it is as readable as the user makes it.
		We strive to keep all the functions simple and intuitive to allow a good readability and a fast scripting.

	- MADCAD is not a parametric software
		a madcad script is just a regular Python script. What makes the madcad files parametric is the ability to change variables and reexecute the script.
		
		So madcad is generative, for implicit geometry definitions, the user has to use a solver before constructing the objects.
	
	- MADCAD is extensible
		As it's a regular script, it's possible to use from simple to complex programming concepts (if/else, loops, tables, functions, classes, module imports, generators, ...)


Features
--------

- surface generation (3D sketch primitives, extrusion, revolution, ...)
- fast boolean operations
- common mesh file format import/export
- kinematic manipulation
- implicit geometry definition through the constraint/solver system
- objects display with high-quality graphics

data types
----------

The most common object types of MADCAD are the following:

math types: 
* vec3    a 3D vector (with fast operations)
* mat3    linear transformation, used for rotations and scaling
* mat4    affine transformation, used for poses, rotations, translations, scaling
* quat    a quaternion, used for rotation (faster and more convenient than matrices for non-repeated operations)

mesh data:
* Mesh		used to represent 3D surfaces
* Web		used to represent 3D lines
* Wire		used to represent only contiguous 3D lines

* kinematic data:
* Solid		each instance constitutes a kinematic undeformable solid
* Kinematic	holds joints and solids for a kinematic structure, with some common mathematical operations

most of the remaining classes are definition elements for kinematics or meshes, see `primitives`, `constraints` and `joints` modules.


Examples
--------

Math operations
***************

	>>> a = vec3(1,2,3)
	>>> O = vec3(0)
	>>> X = vec3(1,0,0)
	>>> normalize(a)
	vec3( 0.267261. 0.534522. 0.801784 )
	>>> 5*X
	vec3(5,0,0)
	>>> cross(a, X)		# cross product  a ^ X
	vec3(0,3,2)
	>>> dot(cross(a, X), X)		# X is orthogonal to its cross product with an other vector
	0
	>>> quat(vec3(0,0,2)) * X		# rotation of the X vector by 2 rad around the Z axis
	vec3( -0.416147. 0.909297. 0 )
	
Primitives
**********

	# define points
	O = vec3(0)
	A = vec3(2,0,0)
	B = vec3(1,2,0)
	C = vec3(0,2,0)
	# create a list of primitives
	line = [
		Line(O, A),          # segment from 0 to A (the direction is important for the surface generation)
		ArcThrough(A, B, C), # arc from A to C, with waypoint B
		Line(C,O),           # segment from C to O
		]

	>>> web(line)	# convert the list of primitives into a Web object, ready for extrusion and so on
	Web( ... )

Solver
******

Suppose that you want to set the Arc tangent to the A and B segments, and fix its radius. It is not easy to guess the precise coordinates for A, B and C for this. You can then specify the constraints to the solver. He will fix that for you.

	csts = [
		Tangent(line[0], line[1], A),   # segment and arc are tangent in A
		Tangent(line[1], line[2], C),   # arc and segment are tangent in C
		Radius(line[1], 1.5),           # radius of arc must be equal to 1.5
		]
	solve(csts, fixed=[0])		# solve the constraints, O is fixed and therefore will not move during the process
	
That's it ! The primitive list can now be converted to Wire or Web with the good shape.	

	>>> A, B, C    # points have been modified inplace
	(vec3(...), vec3(...), vec3(...))

'''
version = '0.18.1'

import os
from .nprint import nprint
curdir = os.path.dirname(__file__)

# computation
from . import (
		# base tools (defines types for the whole library)
		mathutils, mesh, 
		# interdependent functionnalities
		generation, boolean, bevel, primitives, constraints, kinematic, #joints,
		# near-independant modules
		io, hashing, triangulation,
		# parts
		standard,
		settings,
	)
# gui
from . import rendering, displays, text, scheme

# the most common tools, imported to access it directly from madcad
from .mathutils import *
from .mesh import Mesh, Web, Wire, MeshError, web, wire
from .boolean import pierce, difference, union, intersection
from .bevel import chamfer, filet, edgecut, planeoffsets
from .generation import *
from .offseting import *
from .blending import junction, multijunction, blend, blendloop, blendpair, blenditer
from .primitives import isprimitive, Segment, ArcThrough, ArcCentered, ArcTangent, TangentEllipsis, Ellipsis, Circle, Interpolated, Softened
from .constraints import isconstraint, SolveError, Tangent, Distance, Angle, Parallel, Radius, PointOn, OnPlane, solve
from .kinematic import KinematicError, Solid, Joint, Chain, Kinematic
from .reverse import segmentation
from .selection import select
from .io import read, write, cache, cachefunc
from .triangulation import TriangulationError
from .hull import convexhull, convexoutline, horizon
from .hashing import suites

from .scheme import *
from .rendering import Scene, Display, displayable, show, render

from .standard import *




# This file is part of pymadcad,  distributed under license LGPL v3
'''
	The functions here allow to move or extrude meshes along their normals
	
    .. note::
        Any offseting provided here only rely on the vertex and edge normals of the input meshes, so don't expect a nice result if your mesh is too chaotic to have meaningful normals
'''

from .mathutils import *
from .mesh import Mesh, Web, Wire, MeshError, web, wire, mkquad, mktri
from .hashing import edgekey, facekeyo, suites

__all__ = ['thicken', 'inflate', 'inflate_offsets', 'expand']


def inflate_offsets(surface: Mesh, offset: float, method='face') -> '[vec3]':
	''' Displacements vectors for points of a surface we want to inflate.
		
		Parameters:
			offset:
				the distance from the surface to the offset surface. Its meaning depends on `method`
			method:     
				determines if the distance is from the old to the new faces, edges or points
				possible values: `'face', 'edge', 'point'`
	'''
	pnormals = surface.vertexnormals()
	
	# smooth normal offsets laterally when they are closer than `offset`
	outlines = surface.outlines_oriented()
	l = len(pnormals)
	normals = deepcopy(pnormals)
	for i in range(5):	# the number of steps is the diffusion distance through the mesh
		for a,b in outlines:
			d = surface.points[a] - surface.points[b]		# edge direction
			t = cross(pnormals[a]+pnormals[b], d)	# surface tangent normal to the edge
			# contribution stars when the offset points are closer than `offset`
			contrib = 1 - smoothstep(0, offset, length(offset*(pnormals[a]-pnormals[b])+d))
			normals[a] += contrib * 0.5*project(pnormals[b]-pnormals[a], t)
			normals[b] += contrib * 0.5*project(pnormals[a]-pnormals[b], t)
		# renormalize
		for i in range(l):
			pnormals[i] = normals[i] = normalize(normals[i])
	
	# compute offset length depending on the method
	if method == 'face':
		lengths = [inf]*len(pnormals)
		for face in surface.faces:
			fnormal = surface.facenormal(face)
			for p in face:
				lengths[p] = min(lengths[p], 1/dot(pnormals[p], fnormal))
		return typedlist((pnormals[p]*lengths[p]*offset   for p in range(len(pnormals))), dtype=vec3)
	
	elif method == 'edge':
		lengths = [inf]*len(pnormals)
		for edge,enormal in surface.edgenormals().items():
			for p in edge:
				lengths[p] = min(lengths[p], 1/dot(pnormals[p], enormal))
		return typedlist((pnormals[p]*lengths[p]*offset	for p in range(len(pnormals))), dtype=vec3)
		
	elif method == 'point':
		return typedlist((pnormals[p]*offset	for p in range(len(pnormals))), dtype=vec3)

def inflate(surface:Mesh, offset:float, method='face') -> 'Mesh':
	''' Move all points of the surface to make a new one at a certain distance of the last one

		Parameters:
			offset:       the distance from the surface to the offseted surface. its meaning depends on `method`
			method:       determines if the distance is from the old to the new faces, edges or points
        
        Example:
            
            >>> sphere = pierce(
            ...         icosphere(O, 1), 
            ...         brick(min=vec3(0), max=vec3(2)) .flip(),
            ...         )
            >>> inflate(sphere, 0.1)
	'''
	return Mesh(
				typedlist((p+d   for p,d in zip(surface.points, inflate_offsets(surface, offset, method))), dtype=vec3),
				surface.faces,
				surface.tracks,
				surface.groups)

def thicken(surface: Mesh, thickness: float, alignment:float=0, method='face') -> 'Mesh':
	''' Thicken a surface by extruding it, points displacements are made along normal. 

		Parameters:
			thickness:    determines the distance between the two surfaces (can be negative to go the opposite direction to the normal).
			alignment:    specifies which side is the given surface: 0 is for the first, 1 for the second side, 0.5 thicken all apart the given surface.
			method:       determines if the thickness is from the old to the new faces, edges or points
			
        Example:
            
            >>> sphere = pierce(
            ...         icosphere(O, 1), 
            ...         brick(min=vec3(0), max=vec3(2)) .flip(),
            ...         )
            >>> thicken(sphere, 0.1)
	'''
	displts = inflate_offsets(surface, thickness, method)
	
	a = alignment
	b = alignment-1
	m = (	Mesh(
				typedlist((p+d*a  for p,d in zip(surface.points,displts)), dtype=vec3), 
				surface.faces[:], 
				surface.tracks[:], 
				surface.groups)
		+	Mesh(
				typedlist((p+d*b  for p,d in zip(surface.points,displts)), dtype=vec3), 
				surface.faces, 
				surface.tracks, 
				surface.groups[:]
				) .flip() 
		)
	t = len(m.groups)
	l = len(surface.points)
	m.groups.append(None)
	for e in surface.outlines_oriented():
		mkquad(m, (e[0], e[1], e[1]+l, e[0]+l), t)
	return m


def expand(surface: Mesh, offset: float, collapse=True) -> Mesh:
	''' Generate a surface expanding the input mesh on the tangent of the ouline neighboring faces
	
		Parameters:
			offset:		distance from the outline point to the expanded outline points
			collapse:	if True, expanded points leading to crossing edges will collapse into one
			
        Example:
        
            >>> expand(
            ...     revolution(wire([vec3(1,1,0), vec3(0,1,0)]), Axis(O,X), pi),
            ...     0.5)
	'''
	# outline with associated face normals
	pts = surface.points
	edges = {}
	for face in surface.faces:
		for e in ((face[0], face[1]), (face[1], face[2]), (face[2],face[0])):
			if e in edges:	del edges[e]
			else:			edges[(e[1], e[0])] = surface.facenormal(face)
	
	# return the point on tangent for a couple of edges from the frontier
	def tangent(e0, e1):
		mid = axis_midpoint(
				(pts[e0[1]], pts[e0[0]] - pts[e0[1]]), 
				(pts[e1[0]], pts[e1[1]] - pts[e1[0]]),
				)
		d0 = pts[e0[1]] - pts[e0[0]]
		d1 = pts[e1[1]] - pts[e1[0]]
		n0, n1 = edges[e0], edges[e1]
		t = normalize(cross(n0, n1) + NUMPREC*(d0*length(d1)-d1*length(d0)) + NUMPREC**2 * (cross(n0, d0) + cross(n1, d1)))
		if dot(t, cross(n0, d0)) < 0:
			t = -t
		return mid + t * offset
	
	# cross neighbooring normals
	for loop in suites(edges, cut=False):
		assert loop[-1] == loop[0],  "non-manifold input mesh"
		loop.pop()
		# compute the offsets, and remove anticipated overlapping geometries
		extended = [None]*len(loop)
		for i in range(len(loop)):
			# consecutive edges around i-1
			ei0 = (loop[i-2], loop[i-1])
			ei1 = (loop[i-1], loop[i])
			ti = tangent(ei0, ei1)
			if collapse:
				tk = deepcopy(ti)
				weight = 1
				# j is moving to find how much points to gather
				ej0 = ei0
				for j in reversed(range(i-len(loop)+1, i-1)):
					# consecutive edges aroung j
					ej0, ej1 = (loop[j-1], loop[j]), ej0
					tj = tangent(ej0, ej1)
					
					if dot(ti - tj, pts[ei1[0]] - pts[ej0[1]]) <= NUMPREC * length2(pts[ei1[0]] - pts[ej0[1]]):
						tk += tj
						weight += 1
					else:
						break
				# store tangents
				for k in range(j+1, i):
					extended[k] = tk/weight
			else:
				extended[i-1] = ti
		
		# insert the new points
		j = l = len(pts)
		g = len(surface.groups)
		surface.groups.append(None)
		for i in range(len(extended)):
			if extended[i] != extended[i-1]:
				pts.append(extended[i])
				
		# create the faces
		for i in range(len(extended)):
			if extended[i] != extended[i-1]:
				mkquad(surface, (loop[i-1], loop[i], j, (j if j > l else len(pts)) -1), g)
				j += 1
			else:
				mktri(surface, (loop[i-1], loop[i], (j if j > l else len(pts)) -1), g)
	
	return surface


def axis_midpoint(a0: Axis, a1: Axis, x=0.5) -> vec3:
	''' Return the midpoint of two axis. 
		`x` is the blending factor between `a0` and `a1`
		
		- `x = 0` gives the point of `a0` the closest to `a1`
		- `x = 1` gives the point of `a1` the closest to `a0`
	'''
	p0, d0 = a0
	p1, d1 = a1
	if dot(d0,d1)**2 == length2(d0)*length2(d1):
		return mix(p0, p1, 0.5)
	return mix(
		p0 + unproject(project(p1-p0, noproject(d0, d1)), d0),
		p1 + unproject(project(p0-p1, noproject(d1, d0)), d1),
		x)




# This file is part of pymadcad,  distributed under license LGPL v3
__all__ = ['ChainManip', 'KinematicManip', 'scale_solid', 'world_solid']

import warnings
from dataclasses import dataclass
import numpy as np
import numpy.linalg as la
import moderngl as mgl
from PyQt5.QtCore import Qt, QEvent	

from ..common import resourcedir
from ..mathutils import *
from ..mesh import Mesh, Web, Wire, striplist, distance2_pm, typedlist_to_numpy
from .. import settings
from .. import rendering
from .. import scheme
from .. import nprint
from ..displays import BoxDisplay
from ..rendering import Group, displayable
from ..scheme import Scheme, halo_screen
from ..generation import revolution
from ..mesh import web
from .solver import Chain, Kinematic, KinematicError, regularize_grad, structure_state, flatten_state


class SolidDisplay(Group):
	''' Movable `Group` for the rendering pipeline '''
	def __init__(self, scene, solid):
		super().__init__(scene, solid.content, local=solid.pose)
	
	def update(self, scene, solid):
		from .assembly import Solid
		if not isinstance(solid, Solid):	return
		super().update(scene, solid.content)
		self.local = fmat4(solid.pose)
		return True
	
	def stack(self, scene):
		for key,display in self.displays.items():
			if key == 'annotations' and not scene.options['display_annotations'] and not self.selected:
				continue
			for sub,target,priority,func in display.stack(scene):
				yield ((key, *sub), target, priority, func)


class ChainManip(Group):
	''' object to display and interact with a robot in the 3d view

		Attributes:
			chain:  the kinematic chain this display is rendering
			pose:   joints poses in the last rendered frame
			parts:  solids poses in the last rendered frame
			toolcenter (vec3):   current end-solid rotation point in rotation mode, relative to last solid
	'''
	min_colinearity = 1e-2
	max_increment = 0.3
	prec = 1e-6
	
	def __init__(self, scene, chain, pose=None, toolcenter=None):
		super().__init__(scene)
		self.chain = chain
		self.toolcenter = toolcenter or chain.joints[-1].position[1]
		self.pose = pose or chain.default
		self.parts = self.chain.parts(self.pose)
		
		if chain.content:
			for key, solid in enumerate(chain.content):
				if displayable(solid):
					self.displays[key] = scene.display(solid)
		
		scheme, index = kinematic_scheme(chain.joints)
		scheme += kinematic_toolcenter(self.toolcenter)
		self.displays['scheme'] = scene.display(scheme)
		
	# TODO: add content update method
	
	def stack(self, scene):
		''' rendering stack requested by the madcad rendering system '''
		yield ((), 'screen', -1, self.place_solids)
		yield from super().stack(scene)

	def place_solids(self, view):
		world = self.world * self.local
		
		index = {}
		for i, joint in enumerate(self.chain.joints):
			solid = joint.solids[-1]
			index[solid] = i+1
			if display := self.displays.get(solid):
				display.world = world * fmat4(self.parts[i+1])
		
		for space in self.displays['scheme'].spacegens:
			if isinstance(space, (world_solid, scale_solid)) and space.solid in index:
				space.pose = world * fmat4(self.parts[index[space.solid]])
			elif isinstance(space, scheme.halo_screen):
				if view.scene.options['kinematic_manipulation'] == 'rotate':
					space.position = fvec3(self.parts[-1] * self.toolcenter)
				else:
					space.position = fvec3(nan)

	def control(self, view, key, sub, evt):
		''' user event manager, optional part of the madcad rendering system '''
		if evt.type() == QEvent.MouseButtonPress and evt.button() == Qt.LeftButton:
			evt.accept()
			
		if evt.type() == QEvent.MouseButtonPress and evt.button() == Qt.RightButton:
			evt.accept()
			self.panel.setParent(view)
			self.panel.move(evt.pos())
			self.panel.show()
		
		if evt.type() == QEvent.MouseMove and evt.buttons() & Qt.LeftButton:
			evt.accept()
			# put the tool into the view, to handle events
			# TODO: allow changing mode during move
			if sub == ('scheme', index_toolcenter):
				view.tool.append(rendering.Tool(self.move_tool, view, sub, evt))
			else:
				view.tool.append(rendering.Tool(getattr(self, 'move_'+view.scene.options['kinematic_manipulation']), view, sub, evt))
				
	def move_tool(self, dispatcher, view, sub, evt):
		place = mat4(self.world) * self.parts[-1]
		init = place * vec3(self.toolcenter)
		offset = init - view.ptfrom(evt.pos(), init)
		
		while True:
			evt = yield
			# drag
			if evt.type() in (QEvent.MouseMove, QEvent.MouseButtonRelease):
				evt.accept()
				view.update()
				if not (evt.buttons() & Qt.LeftButton):
					break
			
			self.toolcenter = affineInverse(place) * (view.ptfrom(evt.pos(), init) + offset)
	
	def move_joint(self, dispatcher, view, sub, evt):
		
		# find the clicked solid, and joint controled
		if sub[0] == 'scheme':	solid = sub[1]
		else:					solid = sub[0]
		
		# get 3d location of mouse
		click = view.ptat(view.somenear(evt.pos()))
		joint = max(0, solid-1)
		anchor = affineInverse(mat4(self.world * self.local) * self.parts[joint+1]) * vec4(click,1)
		
		while True:
			evt = yield
			# drag
			if evt.type() in (QEvent.MouseMove, QEvent.MouseButtonRelease):
				evt.accept()
				view.update()
				if not (evt.buttons() & Qt.LeftButton):
					break

				model = mat4(view.uniforms['proj'] * view.uniforms['view'] * self.world * self.local) * self.parts[joint]
				direct = affineInverse(self.parts[joint]) * self.parts[joint+1]
				target = qtpos(evt.pos(), view)
				current = model * (direct * anchor)
				move = target - current.xy / current.w

				# solid move directions that can be acheived with this joint
				jac = regularize_grad(self.chain.joints[joint].grad(self.pose[joint]))
				# gradient of the screen position
				jac = np.stack([
					(model * (grad * anchor)).xy / current.w
					for grad in jac])
				# colinearity between the desired move and the gradient directions. it avoids the mechanisme to burst when close to a singularity and also when the mouse move is big
				colinearity = min(1, sum(
							dot(grad, move)**2 / (length2(grad) + length2(move) + self.prec)
							for grad in jac) / self.min_colinearity)
				# combination of the closest directions to the mouse position
				increment = la.solve(jac @ jac.transpose() + np.eye(len(jac))*self.prec, jac @ (move * colinearity))
				self.pose[joint] = self.chain.joints[joint].normalize(
										(np.asarray(self.pose[joint]) 
										+ increment * min(1, self.max_increment / np.abs(increment).max())
									) .clip(*self.chain.joints[joint].bounds))
				self.parts = self.chain.parts(self.pose)

	def move(self, move, jac):
		''' newton method step for moving the mechanism
		'''
		colinearity = min(1, sum(
			np.dot(grad, move)**2 / (normsq(grad) + normsq(move) + self.prec)
			for grad in jac) / self.min_colinearity)
		
		increment = la.solve(jac @ jac.transpose() + np.eye(len(jac))*self.prec, jac @ (move * colinearity))
		
		self.pose = self.chain.normalize(structure_state((
						flatten_state(self.pose)
						+ increment * min(1, self.max_increment / np.abs(increment).max())
						),
						self.pose))
		self.parts = self.chain.parts(self.pose)

	def move_translate(self, dispatcher, view, sub, evt):
		# translate the tool
		clicked = view.ptat(view.somenear(evt.pos()))
		solid = self.parts[-1]
		anchor = affineInverse(mat4(self.world * self.local) * solid) * vec4(clicked,1)
		init_solid = solid
		
		while True:
			evt = yield
			# drag
			if evt.type() in (QEvent.MouseMove, QEvent.MouseButtonRelease):
				evt.accept()
				view.update()
				if not (evt.buttons() & Qt.LeftButton):
					break

				solid = self.parts[-1]
				jac = self.chain.grad(self.pose)
				model = mat4(view.uniforms['proj'] * view.uniforms['view'] * self.world * self.local)
				current_anchor = model * solid * anchor
				target_anchor = qtpos(evt.pos(), view)
				
				self.move(
					move = np.concatenate([
						target_anchor - current_anchor.xy / current_anchor.w,
						np.ravel(mat3(init_solid - solid)),
						]),
					jac = np.stack([
						np.concatenate([
							(model * grad * anchor).xy / current_anchor.w, 
							np.ravel(mat3(grad)), 
							])
						for grad in jac]),
					)

	def move_rotate(self, dispatcher, view, sub, evt):
		# translate the tool
		clicked = view.ptat(view.somenear(evt.pos()))
		solid = self.parts[-1]
		anchor = affineInverse(mat4(self.world * self.local) * solid) * vec4(clicked,1)
		tool = vec4(self.toolcenter,1)
		init_tool = solid * tool
		
		while True:
			evt = yield
			# drag
			if evt.type() in (QEvent.MouseMove, QEvent.MouseButtonRelease):
				evt.accept()
				view.update()
				if not (evt.buttons() & Qt.LeftButton):
					break

				solid = self.parts[-1]
				jac = self.chain.grad(self.pose)
				model = mat4(view.uniforms['proj'] * view.uniforms['view'] * self.world * self.local)
				current_tool = model * solid * tool
				target_anchor = qtpos(evt.pos(), view)
				target_tool = model * init_tool
				
				self.move(
					move = np.concatenate([
						(init_tool - solid * tool).xyz,
						(target_anchor - target_tool.xy/target_tool.w) 
							- (model * solid * normalize(anchor - tool)).xy / current_tool.w,
						]),
					jac = np.stack([
						np.concatenate([
							(grad * tool).xyz, 
							(model * grad * normalize(anchor - tool)).xy / current_tool.w, 
							])
						for grad in jac]),
					)
		

		
class KinematicManip(Group):
	''' 
        Display that holds a kinematic structure and allows the user to move it
        
        Attributes:
            kinemaitc:  the kinematic this display is rendering
            pose:       joints poses in the last rendered frame
            parts:      solids poses in the last rendered frame
            toolcenter:  current solid rotation point in rotation mode, relative to kinematic ground
	'''
	max_increment = 0.3
	prec = 1e-6
	
	move_maxiter = 8
	stay_maxiter = 10000
	move_precision = 1e-4
	stay_precision = 1e-6
	tolerated_precision = 1e-1
	
	def __init__(self, scene, kinematic, pose=None, toolcenter=None):
		super().__init__(scene)
		self.kinematic = kinematic
		self.toolcenter = toolcenter or vec3(0)
		try:
			self.pose = self.kinematic.solve(
				close=pose or self.kinematic.default, 
				maxiter=self.stay_maxiter, 
				precision=self.move_precision, 
				strict=self.tolerated_precision,
				)
		except KinematicError:
			warnings.warn("could not solve kinematic {} for displaying, falling back to infinite tolerance".format(self))
			self.pose = self.kinematic.default
			self.tolerated_precision = inf
		self.parts = self.kinematic.parts(self.pose, precision=self.tolerated_precision)
		
		if self.kinematic.content:
			for key, solid in self.kinematic.content.items():
				if displayable(solid):
					self.displays[key] = scene.display(solid)
		
		scheme, self.index = kinematic_scheme(kinematic.joints)
		scheme += kinematic_toolcenter(self.toolcenter)
		self.displays['scheme'] = scene.display(scheme)
		
	def stack(self, scene):
		''' rendering stack requested by the madcad rendering system '''
		yield ((), 'screen', -1, self.place_solids)
		for item in super().stack(scene):
			if not scene.options['display_annotations'] and not self.selected and len(self.displays) >1 and item[0][0] == 'scheme':
				continue
			yield item

	def place_solids(self, view):
		world = self.world * self.local
		
		for key in self.displays:
			if key in self.parts:
				self.displays[key].world = world * fmat4(self.parts[key])
		
		for space in self.displays['scheme'].spacegens:
			if isinstance(space, (world_solid, scale_solid)) and space.solid in self.parts:
				space.pose = fmat4(self.parts[space.solid])
			elif isinstance(space, scheme.halo_screen):
				if view.scene.options['kinematic_manipulation'] == 'rotate':
					space.position = fvec3(self.toolcenter)
				else:
					space.position = fvec3(nan)

	def control(self, view, key, sub, evt):
		''' user event manager, optional part of the madcad rendering system '''
		# give priority to sub elements
		disp = self.displays
		stack = list(key)
		for i in range(1,len(sub)):
			disp = disp[sub[i-1]]
			disp.control(view, sub[:i], sub[i:], evt)
			if evt.isAccepted(): return
			stack.append(disp)
		
		if evt.type() == QEvent.MouseButtonPress and evt.button() == Qt.LeftButton:
			# accept mouse pressing to tell we are interested in mouse movements
			evt.accept()
		
		if evt.type() == QEvent.MouseMove and evt.buttons() & Qt.LeftButton:
			evt.accept()
			# put the tool into the view, to handle events
			# TODO: allow changing mode during move
			if sub == ('scheme', index_toolcenter):
				view.tool.append(rendering.Tool(self.move_tool, view, sub, evt))
			else:
				view.tool.append(rendering.Tool(getattr(self, 'move_'+view.scene.options['kinematic_manipulation']), view, sub, evt))
	
	def move_tool(self, dispatcher, view, sub, evt):
		place = mat4(self.world)
		init = place * vec3(self.toolcenter)
		offset = init - view.ptfrom(evt.pos(), init)
		
		while True:
			evt = yield
			# drag
			if evt.type() in (QEvent.MouseMove, QEvent.MouseButtonRelease):
				evt.accept()
				view.update()
				if not (evt.buttons() & Qt.LeftButton):
					break
			
			self.toolcenter = affineInverse(place) * (view.ptfrom(evt.pos(), init) + offset)
					
	def move_opt(self, optmove, optjac):
		'''
			newton method step for moving the mechanism
			
			Parameters:
				optmove:  the vector of the desired optional movement
				optjac:   the jacobian of the residuals for the desired optional movement
		'''
		costmove = -self.kinematic.cost_residuals(self.pose)
		# priority factor, lowering the optional move while the closed loop error is big
		priority = self.move_precision / max(self.move_precision, (costmove**2).max(initial=0))
		# colinearity factor, lowering the optional move while it doesn't match the mechanism degrees of freedom
		colinearity = min(1, max(
			np.dot(optgrad, optmove)**2 / (normsq(optgrad) * normsq(optmove) + self.prec**4)
			for optgrad in optjac))
		
		# complete problem to solve
		move = np.hstack([
			optmove * colinearity * priority, 
			costmove,
			])
		jac = np.hstack([
			optjac,
			self.kinematic.cost_jacobian(self.pose).transpose(),
			])
		increment = la.solve(jac @ jac.transpose() + np.eye(len(jac))*self.prec, jac @ move)
		# assemble the new pose and normalize it
		newpose = self.kinematic.normalize(structure_state(
			flatten_state(self.pose)
			+ increment * min(1, self.max_increment / np.abs(increment).max()),
			self.pose))
		
		# try to move
		try:
			self.parts = self.kinematic.parts(self.pose, precision=self.tolerated_precision)
			self.pose = newpose
		except KinematicError:
			pass
	
	def move_joint(self, dispatcher, view, sub, evt):
		# identify the solid clicked
		if sub[0] == 'scheme':  moved = self.index[sub[1]]
		else:                   moved = sub[0]
		if moved == self.kinematic.ground:
			return
		# define the kinematic problem in term of that solid, so we can get a gradient
		kinematic = Kinematic(
			ground = self.kinematic.ground, 
			inputs = self.kinematic.joints,
			outputs = [moved],
			)
		# constants during translation
		clicked = view.ptat(view.somenear(evt.pos()))
		solid = self.parts[moved]
		anchor = affineInverse(mat4(self.world * self.local) * solid) * vec4(clicked,1)
		
		while True:
			evt = yield
			# drag
			if evt.type() in (QEvent.MouseMove, QEvent.MouseButtonRelease):
				evt.accept()
				view.update()
				if not (evt.buttons() & Qt.LeftButton):
					break

				model = mat4(view.uniforms['proj'] * view.uniforms['view'] * self.world * self.local)
				target_anchor = qtpos(evt.pos(), view)
				freejac = kinematic.grad((
					*self.pose, 
					kinematic.joints[-1].inverse(self.parts[moved]),
					))
				
				for i in range(self.move_maxiter):
					solid = self.parts[moved]
					current_anchor = model * solid * anchor
					
					self.move_opt(
						optmove = np.asarray(target_anchor - current_anchor.xy / current_anchor.w),
						optjac = np.stack([
								np.asarray((model * grad * anchor).xy / current_anchor.w)
								for grad in freejac]),
						)
	
	def move_translate(self, dispatcher, view, sub, evt):
		# identify the solid clicked
		if sub[0] == 'scheme':  moved = self.index[sub[1]]
		else:                   moved = sub[0]
		if moved == self.kinematic.ground:
			return
		# define the kinematic problem in term of that solid, so we can get a gradient
		kinematic = Kinematic(
			ground = self.kinematic.ground, 
			inputs = self.kinematic.joints,
			outputs = [moved],
			)
		# constants during translation
		clicked = view.ptat(view.somenear(evt.pos()))
		solid = self.parts[moved]
		anchor = affineInverse(mat4(self.world * self.local) * solid) * vec4(clicked,1)
		init_solid = solid
		
		while True:
			evt = yield
			# drag
			if evt.type() in (QEvent.MouseMove, QEvent.MouseButtonRelease):
				evt.accept()
				view.update()
				if not (evt.buttons() & Qt.LeftButton):
					break

				model = mat4(view.uniforms['proj'] * view.uniforms['view'] * self.world * self.local)
				target_anchor = qtpos(evt.pos(), view)
				jac = kinematic.grad((
					*self.pose, 
					kinematic.joints[-1].inverse(self.parts[moved]),
					))
					
				for i in range(self.move_maxiter):
					solid = self.parts[moved]
					current_anchor = model * solid * anchor
				
					self.move_opt(
						optmove = np.concatenate([
							target_anchor - current_anchor.xy / current_anchor.w,
							np.ravel(mat3(init_solid - solid)),
							]),
						optjac = np.stack([
							np.concatenate([
								(model * grad * anchor).xy / current_anchor.w, 
								np.ravel(mat3(grad)), 
								])
							for grad in jac]),
						)
				
	def move_rotate(self, dispatcher, view, sub, evt):
		# identify the solid clicked
		if sub[0] == 'scheme':  moved = self.index[sub[1]]
		else:                   moved = sub[0]
		if moved == self.kinematic.ground:
			return
		# define the kinematic problem in term of that solid, so we can get a gradient
		kinematic = Kinematic(
			ground = self.kinematic.ground, 
			inputs = self.kinematic.joints,
			outputs = [moved],
			)
		# constants during translation
		clicked = view.ptat(view.somenear(evt.pos()))
		solid = self.parts[moved]
		anchor = affineInverse(mat4(self.world * self.local) * solid) * vec4(clicked,1)
		init_tool = vec4(self.toolcenter, 1)
		tool = affineInverse(solid) * vec4(self.toolcenter, 1)
		
		while True:
			evt = yield
			# drag
			if evt.type() in (QEvent.MouseMove, QEvent.MouseButtonRelease):
				evt.accept()
				view.update()
				if not (evt.buttons() & Qt.LeftButton):
					break

				model = mat4(view.uniforms['proj'] * view.uniforms['view'] * self.world * self.local)
				target_anchor = qtpos(evt.pos(), view)
				target_tool = model * init_tool
				jac = kinematic.grad((
					*self.pose, 
					kinematic.joints[-1].inverse(self.parts[moved]),
					))
				
				for i in range(self.move_maxiter):
					solid = self.parts[moved]
					current_tool = model * solid * tool
				
					self.move_opt(
						optmove = np.concatenate([
							(init_tool - solid * tool).xyz,
							(target_anchor - target_tool.xy/target_tool.w) 
								- (model * solid * normalize(anchor - tool)).xy / current_tool.w,
							]),
						optjac = np.stack([
							np.concatenate([
								(grad * tool).xyz, 
								(model * grad * normalize(anchor - tool)).xy / current_tool.w, 
								])
							for grad in jac]),
						)




@dataclass
class scale_solid:
	''' scheme space scaling around a point in a given solid '''
	solid: object
	center: fvec3
	size: float
	pose: fmat4 = fmat4()
	
	def __call__(self, view):
		m = view.uniforms['view'] * view.uniforms['world'] * self.pose
		e = view.uniforms['proj'] * fvec4(1,1,(m*self.center).z,1)
		e /= e[3]
		return m * translate(self.center) * scale(fvec3(min(self.size, 2 / (e[1]*view.target.height))))

@dataclass
class world_solid:
	solid: object
	pose: fmat4 = fmat4()
	
	def __call__(self, view):
		return view.uniforms['view'] * view.uniforms['world'] * self.pose



index_toolcenter = 10000
	
def kinematic_toolcenter(toolcenter):
	''' create a scheme for drawing the toolcenter in kinematic manipulation '''
	color = settings.colors['annotation']
	angle = 2
	radius = 60
	size = 3
	tend = -sin(angle)*X + cos(angle)*Y
	rend = cos(angle)*X + sin(angle)*Y
	sch = Scheme(
		space=halo_screen(fvec3(toolcenter)), 
		layer=1e-4,
		shader='line',
		track=index_toolcenter,
		)
	sch.add([size*cos(t)*X + size*sin(t)*Y   for t in linrange(0, 2*pi, step=0.2)], color=fvec4(color,1))
	sch.add([radius*cos(t)*X + radius*sin(t)*Y  for t in linrange(0, angle, step=0.1)], color=fvec4(color,1))
	sch.add([radius*cos(t)*X + radius*sin(t)*Y  for t in linrange(angle, 2*pi, step=0.1)], color=fvec4(color,0.2))
	sch.add(
		revolution(web([radius*rend, radius*rend-14*tend-3.6*rend]), Axis(radius*rend, tend), resolution=('div', 8)), 
		shader='fill',
		color=fvec4(color,0.8),
		)
	return sch

def kinematic_scheme(joints) -> '(Scheme, index)':
	''' create a kinematic scheme for the given joints '''
	index = {}
	centers = {}
	for joint in joints:
		if hasattr(joint, 'position'):
			for solid, position in zip(joint.solids, joint.position):
				centers[solid] = centers.get(solid, 0) + vec4(position, 1)
				index[solid] = index.get(solid, len(index))
	for solid, center in centers.items():
		centers[solid] = center.xyz / center.w  if center.w > 1 else None
	
	scheme = Scheme()
	for joint in joints:
		if hasattr(joint, 'position'):
			size = vec2(sum(
				vec2(distance(position, centers[solid]), 1)
				for solid, position in zip(joint.solids, joint.position)
				if centers[solid]))
			if size.y:
				size = 1.5*size.x / size.y
			else:
				size = inf
			
			sch = joint.scheme(index, size, centers[joint.solids[0]], centers[joint.solids[-1]])
			if sch is not NotImplemented:
				scheme += sch
	
	return scheme, {v: k  for k, v in index.items()}

kinematic_color_names = ['annotation', 'schematic']
def kinematic_color(i):
	''' return the scheme color vector for solid `i` in a kinematic '''
	return fvec4(settings.colors[kinematic_color_names[i%2]], 1)



def qtpos(qtpos, view):
	''' convert qt position in the widget to opengl screen coords in range (-1, 1) '''
	return vec2(
		+ (qtpos.x()/view.width() *2 -1),
		- (qtpos.y()/view.height() *2 -1),
		)

def normsq(x):
	x = x.ravel()
	return np.dot(x,x)



# This file is part of pymadcad,  distributed under license LGPL v3

from math import log, exp, floor
from .mathutils import *
from .mesh import typedlist_to_numpy
from .rendering import Display, overrides, writeproperty
from .common import resourcedir
from . import settings
from . import primitives
from PIL import Image
import numpy.core as np
import moderngl as mgl

from PyQt5.QtCore import Qt, QEvent


def shader_wire(scene):
	return scene.ctx.program(
				vertex_shader=open(resourcedir+'/shaders/wire.vert').read(),
				fragment_shader=open(resourcedir+'/shaders/wire.frag').read(),
				)
	
def shader_uniformcolor(scene):
	return scene.ctx.program(
				vertex_shader=open(resourcedir+'/shaders/uniformcolor.vert').read(),
				fragment_shader=open(resourcedir+'/shaders/uniformcolor.frag').read(),
				)


class PointDisplay(Display):
	def __init__(self, scene, position, size=10, color=None):
		self.position = fvec3(position)
		self.size = size
		self.selected = False
		self.color = fvec3(color or settings.colors['line'])
		
		def load(scene):
			img = Image.open(resourcedir+'/textures/point.png')
			texture = scene.ctx.texture(img.size, 1, img.convert('L').tobytes())
			#self.texture = scene.resource('pointtex', load)
			shader = scene.ctx.program(
						vertex_shader=open(resourcedir+'/shaders/pointhalo.vert').read(),
						fragment_shader=open(resourcedir+'/shaders/pointhalo.frag').read(),
						)
			shader['halotex'].value = 0
			ident_shader = scene.ctx.program(
						vertex_shader=open(resourcedir+'/shaders/pointhalo-ident.vert').read(),
						fragment_shader=open(resourcedir+'/shaders/ident.frag').read(),
						)
			#self.shader = scene.resource('pointhalo', load)
			vb = scene.ctx.buffer(np.array([(0,0), (0,1), (1,1), (0,0), (1,1), (1,0)], 'f4'))
			va = scene.ctx.vertex_array(shader, [(vb, '2f', 'v_uv')])
			va_ident = scene.ctx.vertex_array(ident_shader, [(vb, '2f', 'v_uv')])
			return texture, shader, va, ident_shader, va_ident
		
		(	self.texture, 
			self.shader, 
			self.va, 
			self.ident_shader, 
			self.va_ident	) = scene.resource('pointhalo', load)
			
	@property
	def box(self):
		return Box(center=self.position, width=fvec3(0))
	
	def render(self, view):
		self.shader['color'].write(fvec3(settings.display['select_color_line']) if self.selected else self.color)
		self.shader['position'].write(fvec3(self.world * fvec4(self.position,1)))
		self.shader['view'].write(view.uniforms['view'])
		self.shader['proj'].write(view.uniforms['proj'])
		self.shader['ratio'] = (
				self.size / view.width(),
				self.size / view.height(),
				)
		self.texture.use(0)
		self.va.render(mgl.TRIANGLES)

	def identify(self, view):
		self.ident_shader['ident'] = view.identstep(1)
		self.ident_shader['position'].write(fvec3(self.world * fvec4(self.position,1)))
		self.ident_shader['view'].write(view.uniforms['view'])
		self.ident_shader['proj'].write(view.uniforms['proj'])
		self.ident_shader['ratio'] = (
				1.5 * self.size / view.width(),
				1.5 * self.size / view.height(),
				)
		self.va_ident.render(mgl.TRIANGLES)
	
	def stack(self, scene):
		return ( ((), 'ident', 2, self.identify),
				 ((), 'screen', 2, self.render))


class AxisDisplay(Display):
	pattern = [0, 0.25, 0.45, 0.55, 0.75, 1]
	repetitions = 3
	def __init__(self, scene, axis, interval=None, color=None, pose=fmat4(1)):
		self.origin = fvec3(axis[0])
		self.direction = fvec3(axis[1])
		self.interval = interval
		self.color = fvec3(color or settings.colors['line'])
		self.selected = False
		self.box = Box(center=self.origin, width=fvec3(0))
		
		self.shader, self.va, self.ident_shader, self.va_ident = scene.resource('axis', self.load)
	
	def load(self, scene):
		shader = scene.ctx.program(
					vertex_shader=open(resourcedir+'/shaders/axis.vert').read(),
					fragment_shader=open(resourcedir+'/shaders/axis.frag').read(),
					)
		ident_shader = scene.ctx.program(
					vertex_shader=open(resourcedir+'/shaders/axis-ident.vert').read(),
					fragment_shader=open(resourcedir+'/shaders/ident.frag').read(),
					)
		pts = []
		for i in range(-1, self.repetitions+1):
			for pt in self.pattern:
				if i == -1:					alpha = pt
				elif i == self.repetitions:	alpha = 1-pt
				else:						alpha = 1
				pts.append(((pt+i)/self.repetitions, alpha))
		vb = scene.ctx.buffer(np.array(pts, 'f4'))
		va = scene.ctx.vertex_array(shader, [(vb, 'f f', 'v_absciss', 'v_alpha')])
		va_ident = scene.ctx.vertex_array(ident_shader, [(vb, 'f 12x', 'v_absciss')])
		return shader, va, ident_shader, va_ident
	
	def _disp_interval(self, view):
		if self.interval:	return self.interval
		else:
			size = -view.uniforms['view'][3][2]/6
			return (-0.5*size, size)
	
	def render(self, view):
		self.shader['projview'].write(view.uniforms['projview'])
		self.shader['world'].write(self.world)
		self.shader['origin'].write(self.origin)
		self.shader['direction'].write(self.direction)
		self.shader['interval'] = self._disp_interval(view)
		self.shader['color'].write(fvec3(settings.display['select_color_line']) if self.selected else self.color)
		self.va.render(mgl.LINES)
	
	def identify(self, view):
		self.ident_shader['projview'].write(view.uniforms['projview'])
		self.ident_shader['world'].write(self.world)
		self.ident_shader['origin'].write(self.origin)
		self.ident_shader['direction'].write(self.direction)
		self.ident_shader['interval'] = self._disp_interval(view)
		self.ident_shader['ident'] = view.identstep(1)
		self.va_ident.render(mgl.LINES)
	
	def stack(self, scene):
		return ( ((), 'ident', 2, self.identify),
				 ((), 'screen', 2, self.render))

def npboundingbox(points):
	''' boundingbox for numpy arrays of points on the 3 first components '''
	return Box(
				fvec3([float(np.min(points[:,i])) for i in range(3)]),
				fvec3([float(np.max(points[:,i])) for i in range(3)]),
				)

class AnnotationDisplay(Display):
	def __init__(self, scene, points, color):
		self.color = fvec3(color or settings.colors['annotation'])
		self.selected = False
		self.box = npboundingbox(points)
		# load shader
		def load(scene):
			return scene.ctx.program(
						vertex_shader=open(resourcedir+'/shaders/annotation.vert').read(),
						fragment_shader=open(resourcedir+'/shaders/annotation.frag').read(),
						)
		self.shader = scene.resource('shader_annotation', load)
		# allocate buffers
		self.vb_pts = scene.ctx.buffer(points)
		self.va = scene.ctx.vertex_array(self.shader, [(self.vb_pts, '3f f', 'v_position', 'v_alpha')])
		self.va_ident = scene.ctx.vertex_array(scene.resource('shader_ident'), [(self.vb_pts, '3f 4x', 'v_position')])
		
	def __del__(self):
		self.va.release()
		self.va_ident.release()
		self.vb_pts.release()
	
	@staticmethod
	def buff_ptsalpha(points, alpha):
		return np.hstack((
				np.array([tuple(p) for p in points], 'f4'), 
				np.array(alpha, 'f4', ndmin=2).transpose(),
				))

	def render(self, view):
		self.shader['proj'].write(view.uniforms['proj'])
		self.shader['view'].write(view.uniforms['view'] * self.world)
		self.shader['color'].write(self.color if not self.selected else fvec3(settings.display['select_color_line']))
		self.va.render(mgl.LINES)
	
	def identify(self, view):
		shader = view.scene.resource('shader_ident')
		shader['proj'].write(view.uniforms['proj'])
		shader['view'].write(view.uniforms['view'] * self.world)
		shader['ident'] = view.identstep(1)
		self.va_ident.render(mgl.LINES)
		
	def stack(self, scene):
		return ( ((), 'ident', 2, self.identify),
				 ((), 'screen', 2, self.render)) 


class BoxDisplay(AnnotationDisplay):
	def __init__(self, scene, box, color=None):
		# place points
		x,y,z = box.width
		c = 0.1*min(x,y,z)	# corner
		o = 0.4 # wire alpha
		pts = np.array([
			# corner
			(0,0,0,1),(c,0,0,1),
			(0,0,0,1),(0,c,0,1),
			(0,0,0,1),(0,0,2*c,1),
			
			# box wire
			(0,0,0,o),(x,0,0,o),
			(0,0,0,o),(0,y,0,o),
			(0,0,0,o),(0,0,z,o),
			
			(x,0,0,o),(x,y,0,o),
			(x,0,0,o),(x,0,z,o),
			(0,y,0,o),(x,y,0,o),
			(0,y,0,o),(0,y,z,o),
			(0,0,z,o),(x,0,z,o),
			(0,0,z,o),(0,y,z,o),
			
			(x,y,0,o),(x,y,z,o),
			(x,0,z,o),(x,y,z,o),
			(0,y,z,o),(x,y,z,o),
			], dtype='f4')
		pts += (*box.min, 0)
		self.textplace = box.min

		super().__init__(scene, pts, color)

from .mathutils import norminf, length2

class SolidDisplay(Display):
	''' Display render Meshes '''
	def __init__(self, scene, positions, normals, faces, lines, idents, color=None):
		self.box = npboundingbox(positions)
		self.options = scene.options
		
		color = fvec3(color or settings.colors['surface'])
		line = (	(length(settings.colors['line']) + dot(color-settings.colors['surface'], settings.colors['line']-settings.colors['surface']))
					* normalize(color + 1e-6)  )
		#if length(s['line_color']) > length(color)
		reflect = normalize(color + 1e-6) * settings.display['solid_reflectivity']
		
		self.vertices = Vertices(scene.ctx, positions, idents)
		self.disp_faces = FacesDisplay(scene, self.vertices, normals, faces, color=color, reflect=reflect, layer=0)
		self.disp_ghost = GhostDisplay(scene, self.vertices, normals, faces, color=line, layer=0)
		self.disp_groups = LinesDisplay(scene, self.vertices, lines, color=line, alpha=1, layer=-2e-6)
		self.disp_points = PointsDisplay(scene, self.vertices, range(len(positions)), layer=-3e-6)
		wire = []
		for f in faces:
			wire.append((f[0], f[1]))
			wire.append((f[1], f[2]))
			wire.append((f[2], f[0]))
		self.disp_wire = LinesDisplay(scene, self.vertices, wire, color=line, alpha=0.3, layer=-1e-6)
		
	def stack(self, scene):
		yield ((), 'screen', -1, self.vertices.prerender)
		if self.options['display_faces']:	
			yield ((), 'screen', 0, self.disp_faces.render)
			yield ((), 'ident', 0, self.disp_faces.identify)
		else:								
			yield ((), 'screen', 1, self.disp_ghost.render)
			yield ((), 'ident', 0, self.disp_ghost.identify)
		if self.options['display_groups']:	yield ((), 'screen', 1, self.disp_groups.render)
		if self.options['display_points']:	yield ((), 'screen', 2, self.disp_points.render)
		if self.options['display_wire']:	yield ((), 'screen', 2, self.disp_wire.render)
	
	@property
	def world(self):	return self.vertices.world
	@world.setter
	def world(self, value):	self.vertices.world = value

	#def control(self, view, key, sub, evt):
		#if evt.type() == QEvent.MouseButtonRelease and evt.button() == Qt.LeftButton:
			#sub = sub[0]
			#flags, idents = self.vertices.flags, self.vertices.idents
			#for i in range(len(idents)):
				#flags[i] ^= idents[i] == sub
			#self.vertices.flags_updated = True
			#view.update()
			#evt.accept()
	

class WebDisplay(Display):
	''' Display to render Webs '''
	def __init__(self, scene, positions, lines, points, idents, color=None):
		self.box = npboundingbox(positions)
		self.options = scene.options
		color = color or settings.colors['line']
		self.vertices = Vertices(scene.ctx, positions, idents)
		self.disp_edges = LinesDisplay(scene, self.vertices, lines, color=color, alpha=1, layer=-2e-6)
		self.disp_groups = PointsDisplay(scene, self.vertices, points, layer=-3e-6)
		self.disp_points = PointsDisplay(scene, self.vertices, range(len(positions)), layer=-1e-6)
		

	def stack(self, scene):
		yield ((), 'screen', -1, self.vertices.prerender)
		if self.options['display_groups']:		yield ((), 'screen', 2, self.disp_groups.render)
		if self.options['display_points']:		yield ((), 'screen', 2, self.disp_points.render)
		yield ((), 'screen', 1, self.disp_edges.render)
		yield ((), 'ident', 1, self.disp_edges.identify)
	
	@property
	def world(self):	return self.vertices.world
	@world.setter
	def world(self, value):	self.vertices.world = value

	#def control(self, view, key, sub, evt):
		#if evt.type() == QEvent.MouseButtonRelease and evt.button() == Qt.LeftButton:
			#sub = sub[0]
			#flags, idents = self.vertices.flags, self.vertices.idents
			#for i in range(len(idents)):
				#flags[i] ^= idents[i] == sub
			#self.vertices.flags_updated = True
			#view.update()
			#evt.accept()


class Vertices(object):
	''' convenient class to share vertices between SolidDisplay, WebDisplay, PointsDisplay '''
	def __init__(self, ctx, positions, idents):
		self.idents = idents
		self.nident = int(max(idents))+1
		self.flags = np.zeros(len(positions), dtype='u1')
		self.flags_updated = False
		assert len(idents) == len(positions)
		self.vb_positions = ctx.buffer(np.asarray(positions, dtype='f4', order='C'))
		self.vb_idents = ctx.buffer(np.asarray(idents, dtype='u2', order='C'))
		self.vb_flags = self.vb_flags = ctx.buffer(self.flags, dynamic=True)
		self.world = fmat4(1)
		
	def __del__(self):
		self.vb_positions.release()
		self.vb_idents.release()
		self.vb_flags.release()
		
	def prerender(self, view):
		if self.flags_updated:
			self.vb_flags.write(self.flags)
			self.flags_updated = False
	
	def selectsub(self, sub):
		for i,id in enumerate(self.idents):
			self.flags[i] ^= id == sub
		self.flags_updated = True
			


class FacesDisplay:
	def __init__(self, scene, vertices, normals, faces, color, reflect, layer=0):
		self.color = color
		self.layer = layer
		self.reflect = reflect
		self.vertices = vertices
	
		# load the skybox texture
		def load(scene):
			img = Image.open(resourcedir+'/textures/'+settings.display['solid_reflect'])
			return scene.ctx.texture(img.size, 3, img.tobytes())
		self.reflectmap = scene.resource('skybox', load)
		
		# load the shader
		def load(scene):
			shader = scene.ctx.program(
						vertex_shader=open(resourcedir+'/shaders/solid.vert').read(),
						fragment_shader=open(resourcedir+'/shaders/solid.frag').read(),
						)
			# setup some uniforms
			shader['reflectmap'] = 0
			return shader
		self.shader = scene.resource('shader_solid', load)
		self.ident_shader = scene.resource('shader_subident')
		# allocate buffers
		if faces is not None and len(faces) and vertices.vb_positions:
			self.vb_faces = scene.ctx.buffer(np.asarray(faces, 'u4', order='C'))
			self.vb_normals = scene.ctx.buffer(np.asarray(normals, 'f4', order='C'))
			self.va = scene.ctx.vertex_array(
					self.shader, 
					[	(vertices.vb_positions, '3f', 'v_position'), 
						(self.vb_normals, '3f', 'v_normal'),
						(vertices.vb_flags, 'u1', 'v_flags')],
					self.vb_faces,
					)
			
			self.va_ident = scene.ctx.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_faces,
					)
		else:
			self.va = None
			
	def __del__(self):
		if self.va:
			self.va.release()
			self.va_ident.release()
			self.vb_faces.release()
			self.vb_normals.release()
	
	def render(self, view):
		if self.va:
			# setup uniforms
			self.shader['select_color'].write(settings.display['select_color_face'])
			self.shader['min_color'].write(self.color * settings.display['solid_color_side'])
			self.shader['max_color'].write(self.color * settings.display['solid_color_front'])
			self.shader['refl_color'].write(self.reflect)
			self.shader['layer'] = self.layer
			self.shader['world'].write(self.vertices.world)
			self.shader['view'].write(view.uniforms['view'])
			self.shader['proj'].write(view.uniforms['proj'])
			# render on self.context
			self.reflectmap.use(0)
			self.va.render(mgl.TRIANGLES)
	
	def identify(self, view):
		if self.va:
			self.ident_shader['layer'] = self.layer
			self.ident_shader['start_ident'] = view.identstep(self.vertices.nident)
			self.ident_shader['view'].write(view.uniforms['view'] * self.vertices.world)
			self.ident_shader['proj'].write(view.uniforms['proj'])
			# render on self.context
			self.va_ident.render(mgl.TRIANGLES)

class GhostDisplay:
	def __init__(self, scene, vertices, normals, faces, color, layer=0):
		self.color = color
		self.layer = layer
		self.vertices = vertices
		
		# load the shader
		def load(scene):
			shader = scene.ctx.program(
						vertex_shader=open(resourcedir+'/shaders/solid.vert').read(),
						fragment_shader=open(resourcedir+'/shaders/ghost.frag').read(),
						)
			# setup some uniforms
			return shader
		self.shader = scene.resource('shader_ghost', load)
		self.ident_shader = scene.resource('shader_subident')
		# allocate buffers
		if faces is not None and len(faces) and vertices.vb_positions:
			self.vb_faces = scene.ctx.buffer(np.asarray(faces, 'u4', order='C'))
			self.vb_normals = scene.ctx.buffer(np.asarray(normals, 'f4', order='C'))
			self.va = scene.ctx.vertex_array(
					self.shader, 
					[	(vertices.vb_positions, '3f', 'v_position'), 
						(self.vb_normals, '3f', 'v_normal'),
						(vertices.vb_flags, 'u1', 'v_flags')],
					self.vb_faces,
					)
			
			self.va_ident = scene.ctx.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_faces,
					)
		else:
			self.va = None
			
	def __del__(self):
		if self.va:
			self.va.release()
			self.va_ident.release()
			self.vb_faces.release()
			self.vb_normals.release()
	
	def render(self, view):
		if self.va:
			# setup uniforms
			self.shader['select_color'].write(settings.display['select_color_line'])
			self.shader['normal_color'].write(self.color)
			self.shader['world'].write(self.vertices.world)
			self.shader['view'].write(view.uniforms['view'])
			self.shader['proj'].write(view.uniforms['proj'])
			self.shader['layer'] = self.layer
			view.scene.ctx.disable(mgl.DEPTH_TEST)
			# render on self.context
			self.va.render(mgl.TRIANGLES)
			view.scene.ctx.enable(mgl.DEPTH_TEST)
	
	def identify(self, view):
		if self.va:
			self.ident_shader['start_ident'] = view.identstep(self.vertices.nident)
			self.ident_shader['view'].write(view.uniforms['view'] * self.vertices.world)
			self.ident_shader['proj'].write(view.uniforms['proj'])
			self.ident_shader['layer'] = self.layer
			# render on self.context
			self.va_ident.render(mgl.TRIANGLES)
	
class LinesDisplay:
	def __init__(self, scene, vertices, lines, color, alpha=1, layer=0):
		self.layer = layer
		self.color = fvec4(fvec3(color), alpha)
		self.select_color = fvec4(settings.display['select_color_line'], alpha)
		self.vertices = vertices
		
		# load the line shader
		self.shader = scene.resource('shader_wire', shader_wire)
		self.ident_shader = scene.resource('shader_subident')
		if lines is not None and len(lines) and vertices.vb_positions:
			# allocate buffers
			self.vb_lines = scene.ctx.buffer(np.asarray(lines, dtype='u4', order='C'))
			self.va = scene.ctx.vertex_array(
						self.shader,
						[	(vertices.vb_positions, '3f', 'v_position'),
							(vertices.vb_flags, 'u1', 'v_flags')],
						self.vb_lines,
						)
			self.va_ident = scene.ctx.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_lines,
					)
		else:
			self.va = None
			
	def __del__(self):
		if self.va:
			self.va.release()
			self.va_ident.release()
			self.vb_lines.release()
	
	def render(self, view):
		if self.va:
			self.shader['color'].write(self.color)
			self.shader['select_color'].write(self.select_color)
			self.shader['view'].write(view.uniforms['view'] * self.vertices.world)
			self.shader['proj'].write(view.uniforms['proj'])
			self.shader['layer'] = self.layer
			self.va.render(mgl.LINES)
		
	def identify(self, view):
		if self.va:
			self.ident_shader['start_ident'] = view.identstep(self.vertices.nident)
			self.ident_shader['view'].write(view.uniforms['view'] * self.vertices.world)
			self.ident_shader['proj'].write(view.uniforms['proj'])
			self.ident_shader['layer'] = self.layer
			self.va_ident.render(mgl.LINES)
		
class PointsDisplay:
	def __init__(self, scene, vertices, indices=None, color=None, ptsize=3, layer=0):
		self.color = fvec4(color or settings.colors['point'], 1)
		self.select_color = fvec4(settings.display['select_color_line'], 1)
		self.ptsize = ptsize
		self.layer = layer
		self.vertices = vertices
		
		# load the line shader
		self.shader = scene.resource('shader_wire', shader_wire)
		self.ident_shader = scene.resource('shader_subident')
		# allocate GPU objects
		if indices is not None and len(indices) and vertices.vb_positions:
			self.vb_indices = scene.ctx.buffer(np.asarray(indices, dtype='u4', order='C'))
			self.va = scene.ctx.vertex_array(
						self.shader,
						[	(vertices.vb_positions, '3f', 'v_position'),
							(vertices.vb_flags, 'u1', 'v_flags')],
						self.vb_indices,
						)
			self.va_ident = scene.ctx.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_indices,
					)
		else:
			self.va = None
			
	def __del__(self):
		if self.va:
			self.va.release()
			self.va_ident.release()
			self.vb_indices.release()
	
	def render(self, view):
		if self.va:
			self.shader['layer'] = self.layer
			self.shader['color'].write(self.color)
			self.shader['select_color'].write(self.select_color)
			self.shader['view'].write(view.uniforms['view'] * self.vertices.world)
			self.shader['proj'].write(view.uniforms['proj'])
			view.scene.ctx.point_size = self.ptsize
			self.va.render(mgl.POINTS)
	
	def identify(self, view):
		if self.va:
			scene.subident_shader['layer'] = self.layer
			scene.subident_shader['start_ident'] = view.identstep(self.vertices.nident)
			scene.subident_shader['view'].write(view.uniforms['view'] * self.vertices.world)
			scene.subident_shader['proj'].write(view.uniforms['proj'])
			view.ctx.point_size = self.ptsize
			self.va_ident.render(mgl.POINTS)

def digitfit(n):
	''' return the number of zeros in the digital representation of n '''
	s = 0
	while n:
		if not n%10:	s += 1
		n //=10
	return s

class GridDisplay(Display):
	''' display a grid clipped to the view, helping to appreciate distances
		The grid display distances in the plane of the `center` point
		
		:unit:      factor on the distance between dots
		:color:     base color of the grid, with an alpha channel
		:contrast:  contrast between the smallest division and the biggest
	'''
	def __init__(self, scene, center, unit=1, color=None, contrast=1.8):
		def load(scene):
			n = 101
			h = n//2
			pts = np.empty((n,n), dtype='f4, f4, f2')
			for i in range(n):
				for j in range(n):
					pts[i,j] = (i-h, j-h, (1+min(digitfit(i), digitfit(j))))
			
			shader = scene.ctx.program(
						vertex_shader=open(resourcedir+'/shaders/viewgrid.vert').read(),
						fragment_shader=open(resourcedir+'/shaders/viewgrid.frag').read(),
						)
			vb = scene.ctx.buffer(pts)
			va = scene.ctx.vertex_array(shader, [(vb, '2f4 f2', 'v_position', 'v_opacity')])
			return shader, va
		self.shader, self.va = scene.resource('viewgrid', load)
		self.unit = unit
		self.center = fvec3(center or 0)
		self.color = fvec4(color) if color else fvec4(settings.colors['point'],1)
		self.contrast = contrast
	
	def render(self, view):
		center = fvec3( view.uniforms['view'] * self.world * fvec4(self.center,1) )
		if not center.z < 0:	return
		zlog = log(-center.z)/log(10)
		sizelog = floor(zlog)
		
		view.scene.ctx.point_size = 1/400 * view.fb_screen.height
		self.shader['color'].write(fvec4(
				self.color.rgb, 
				self.color.a * exp(self.contrast*(-1+sizelog-zlog)),
				))
		self.shader['size'] = self.unit * 10**(sizelog-1)
		self.shader['centerdist'] = -center.z
		self.shader['proj'].write(view.uniforms['proj'])
		self.shader['contrast'] = self.contrast
		self.va.render(mgl.POINTS)
	
	def stack(self, scene):
		return ((), 'screen', 3, self.render),
		

class SplineDisplay(Display):
	''' display for spline curve, with handles around'''
	def __init__(self, scene, handles, curve, color=None):
		self.color = color or fvec4(settings.colors['line'], 1)
		self.color_handles = fvec4(settings.colors['annotation'], 0.6)
		self.box = npboundingbox(handles)
		ctx = scene.ctx
		self.vb_handles = ctx.buffer(handles)
		self.vb_curve = ctx.buffer(curve)
		
		self.shader = scene.resource('shader_uniformcolor', shader_uniformcolor)
		self.va_handles = ctx.vertex_array(self.shader, [(self.vb_handles, '3f4', 'v_position')])
		self.va_curve = ctx.vertex_array(self.shader, [(self.vb_curve, '3f4', 'v_position')])
		
		self.shader_ident = scene.resource('shader_ident')
		self.va_ident = ctx.vertex_array(self.shader_ident, [(self.vb_curve, '3f4', 'v_position')])
		
	def __del__(self):
		self.va_handles.release()
		self.va_curve.release()
		self.va_ident.release()
		self.vb_handles.release()
		self.vb_curve.release()
		
	def render(self, view):
		self.shader['view'].write(view.uniforms['view'] * self.world)
		self.shader['proj'].write(view.uniforms['proj'])
		view.scene.ctx.point_size = 4
		
		self.shader['layer'] = -2e-6
		self.shader['color'].write(self.color_handles if not self.selected else fvec4(settings.display['select_color_line'],self.color_handles[3]))
		self.va_handles.render(mgl.POINTS)
		self.va_handles.render(mgl.LINE_STRIP)
		
		self.shader['layer'] = -1e-6
		self.shader['color'].write(self.color if not self.selected else fvec4(settings.display['select_color_line'],self.color[3]))
		self.va_curve.render(mgl.LINE_STRIP)
		
	def identify(self, view):
		self.shader_ident['ident'] = view.identstep(1)
		self.shader_ident['view'].write(view.uniforms['view'] * self.world)
		self.shader_ident['proj'].write(view.uniforms['proj'])
		self.va_ident.render(mgl.LINE_STRIP)
		
	def stack(self, scene):
		return (	((), 'screen', 1, self.render),
					((), 'ident', 1, self.identify)	)
					
					
					
class VoxelDisplay(Display):
	''' display a voxel as a scalar field.
		its opacity is computed using its density as absorbance.
		
		Parameters:
			
			voxel:	a numpy array with dimension 3, no specific value range is required
			
		Attributes:
		
			space:	transformation matrix from voxel normalized coordinates (position in the array as range 0-1) to local coordinates
			world:	transformation matrix from local to world coordinates
			value_range:	
				
				tuple `(min, max)`  giving the min and max values to display
				- values below min will be transparent
				- values above max will be fully opaq
				
			color_range:	tuple `(fvec4, fvec4)` of colors matching the value range bounds
		
		value examples:
		
			* value_range
			
				```
				(0.4, 0.9)  # erase the lower value and saturate the maxium ones
				```
				
			* color_range
			
				```
				(fvec4(0,0,1,1), fvec4(0,1,0,1))	 # blue-green
				(fvec4(0,0.5,1,1), fvec4(1,1,0,1))   # lightblue-yellow
				(fvec4(1,0,0.3,1), fvec4(1,1,0,1))   # red-yellow
				(fvec4(0,0,1,1), fvec4(1,0.5,0.1,1)) # blue-orange
				(fvec4(0,0,1,1), fvec4(1,0.1,0.1,1)) # blue-red
				(fvec4(0,0,0,1), fvec4(1,1,1,1))     # black-white
				(fvec4(1,1,1,1), fvec4(1,1,1,1))     # white-white
				```
	'''
	def __init__(self, scene, voxel: 'ndarray', space: 'fmat4', 
				value=(0, 1), 
				color=(fvec4(0,0,1,1), fvec4(0,1,0,1))):
		
		self.voxel = scene.ctx.texture3d(
			voxel.shape, 
			1, 
			voxel.transpose(2,1,0).copy().astype('f2'),  # axis permutation for opengl
			dtype='f2',
			)
		self.voxel.repeat_x = False
		self.voxel.repeat_y = False
		self.voxel.repeat_z = False
		self.space = space
		self.value_range = value
		self.color_range = color

		def load(scene):
			from . import generation
			
			# load shader
			shader = scene.ctx.program(
						vertex_shader=open(resourcedir+'/shaders/voxel.vert').read(),
						fragment_shader=open(resourcedir+'/shaders/voxel.frag').read(),
						)
			# load vertex buffer for the brick
			brick = generation.brick(min=vec3(0), max=vec3(1)) .flip()
			pts = typedlist(vec3)
			for face in brick.faces:
				pts.extend(brick.facepoints(face))
			vb = scene.ctx.buffer(typedlist_to_numpy(pts, 'f4'))
			
			return shader, vb
		
		self.shader, self.vb = scene.resource('shader_voxel', load)
		self.va = scene.ctx.vertex_array(
				self.shader,
				[(self.vb, '3f', 'v_position')],
				)
				
	def __del__(self):
		self.va.release()
		self.voxel.release()

	def render(self, view):
		view.scene.ctx.enable(mgl.DEPTH_TEST | mgl.CULL_FACE)
		self.shader['value_min'] = self.value_range[0]
		self.shader['value_max'] = self.value_range[1]
		self.shader['color_min'].write(self.color_range[0])
		self.shader['color_max'].write(self.color_range[1])
		self.shader['view'].write(view.uniforms['view'] * self.world * self.space)
		self.shader['proj'].write(view.uniforms['proj'])
		self.voxel.use(0)
		self.va.render(mgl.TRIANGLES)
		
	def stack(self, scene):
		return ((), 'screen', 10, self.render),




def tupledisplay(scene, t):
	if primitives.isaxis(t):	return AxisDisplay(scene, t)
	# if not found: empty display
	return Display()

overrides.update({
	vec3:   PointDisplay,
	tuple:  tupledisplay,
	Box:    BoxDisplay,
	})

	



'''	 The settings module holds dictionaries each aspect of the madcad library.

dictionaries:
	:primitives:  default settings for mesh and primitive operations
	:display:	visual settings to display the 3D objects
	:scene:		for what and how to display in the 3D scene
	:controls:	preferences for the controls of the Scene widget
'''

from math import pi, ceil, floor, sqrt
from glm import fvec3, fvec4, normalize
import sys, os, yaml
from os.path import dirname, exists

# settings for all Display objects and for the scene rendering
display = {
	'field_of_view': pi/6,  # camera field of view for scene views
	'view_font_size': 8,  # font size for annotations
	'joint_size': 50, # maximum pixel size of joints
	'sharp_angle': pi/6, # above this threshold angles between mesh triangles are considered sharp, below they are rendered softened
	
	'background_color': fvec3(0,0,0),  # color of the view background
	'select_color_face': fvec3(0.01, 0.05, 0.03), # color for selected surfaces
	'select_color_line': fvec3(0.5, 1, 0.6), # color for selected lines
	'highlight_color': fvec3(0.1, 0.2, 0.2),
	
	'solid_color_front': 1., # surface intensity for portions of surface parallel to the view direction
	'solid_color_side': 0.2, # surface intensity for portions of surface orthogonal to the view direction
	'solid_reflectivity': 4, # intensity of surface reflections
	'solid_reflect': 'skybox-white.png', # skybox texture for reflects
	'line_width': 1.0, # pixel width of edges, may work or not depending on openGL implementation
	
	'system_theme': True, # adapt madcad colors to the current desktop theme
	}

# theme colors for objects in the scene
colors = {
	# display elements
	
	# mesh surface
	'surface': fvec3(0.2, 0.2, 0.2),
	# mesh and primitives lines
	'line': fvec3(0.9, 0.9, 0.9),
	# mesh and primitives points
	'point': fvec3(0.9, 0.9, 0.9),
	# annotations like dimensions and constraints
	'annotation': fvec3(0.2, 0.7, 1),
	# schematics like kinematic schemes
	'schematic': fvec3(0.3, 0.8, 1),
	
	# color for standard parts
	
	# screws and nults
	'bolt': fvec3(0.2),
	# gears and transmission parts
	'gear': fvec3(0.2, 0.3, 0.4),
	# bearing and guiding parts
	'bearing': fvec3(0.5,0.4,0.35),
	'bearing_cage': fvec3(0.3,0.2,0),
	# deformable parts like springs
	'spring': fvec3(0.2),
	# circulating elements like balls in bearings
	'circulating': fvec3(0,0.1,0.2),
	}

# initial settings for a scene, each scene will duplicate this and will be able to modify them at runtime
# this dictionnary represent the initial settings of the user
scene = {
	'projection': 'Perspective',
	
	'display_faces': True,
	'display_groups': True,
	'display_points': False,
	'display_wire': False,
	'display_grid': True,
	'display_annotations': True,
	'surface_shading': True,
	'kinematic_manipulation': 'joint',
	
	'lock_solids': True,
	
	'debug_points': False,
	'debug_faces': False,
	'debug_groups': False,
	}
		
controls = {
	'navigation': 'Turntable',
	'snap_dist': 10,	# pixel distance to click items
	}

resolution = ['rad', pi/16]  # maximum angle for discretizing curves, see `curve_resolution()`


# get configuration directory depending on OS
if sys.platform == 'win32':
	home = os.getenv('USERPROFILE')
	configdir = home+'/AppData/Local'
else:
	home = os.getenv('HOME')
	configdir = home+'/.config'

config = configdir+'/madcad/pymadcad.yaml'
settings = {'display':display, 'scene':scene, 'controls':controls, 'resolution':resolution}


def install():
	''' Create and fill the config directory if not already existing '''
	if not exists(config):
		os.makedirs(dirname(config), exist_ok=True)
		dump()
		
def clean():
	''' Delete the default configuration file '''
	os.rm(config)

def load(file=None):
	''' Load the settings directly in this module, from the specified file or the default one '''
	if not file:	file = config
	if isinstance(file, str):	file = open(file, 'r')
	changes = yaml.safe_load(file)
	def update(dst, src):
		for key in dst:
			if key in src:
				if isinstance(dst[key], dict) and isinstance(src[key], dict):	
					update(dst[key], src[key])
				elif isinstance(dst[key], fvec3):	dst[key] = fvec3(src[key])
				elif isinstance(dst[key], fvec4):	dst[key] = fvec4(src[key])
				else:
					dst[key] = src[key]
	update(settings, changes)

def dump(file=None):
	''' Load the current settings into the specified file or to the default one '''
	if not file:	file = config
	if isinstance(file, str):	file = open(file, 'w')
	yaml.add_representer(fvec3, lambda dumper, data: dumper.represent_list(round(f,3) for f in data))
	yaml.add_representer(fvec4, lambda dumper, data: dumper.represent_list(round(f,3) for f in data))
	file.write(yaml.dump(settings, default_flow_style=None, width=40, indent=4))




def curve_resolution(length, angle, param=None):
	''' Return the subdivision number for a curve, using the given or setting specification

		:length:  is the curvilign length of the curve
		:angle:   is the integral of the absolute curvature (total rotation angle)
	'''
	kind, prec = param or resolution
	if kind == 'div':
		res = prec
	elif kind == 'm':
		res = ceil(length / prec)
	elif kind == 'rad':
		res = floor(angle / prec)
	elif kind == 'radm':
		res = floor(length*angle / prec)
	elif kind == 'sqradm':
		res = floor(sqrt(length*angle) / prec)
	else:
		raise ValueError("unknown type for round_limit: {}".format(repr(kind)))
	res = max(res, floor(angle * 2/pi))
	return res

def getparam(levels: list, key):
	''' Get the first found value for key through the given dictionnaries.
		Dictionnaries are tested successively until the matching value is found. If no value is found, None is returned
	'''
	for d in levels:
		if d is not None:
			if key in d:	return d[key]
	return None

def use_qt_colors():
	''' Set the color settings to fit the current system colors '''
	from .mathutils import fvec3, mix, distance
	from PyQt5.QtWidgets import QApplication
	palette = QApplication.instance().palette()
	def qtc(role):
		''' Convert a QColor or QPalette role to fvec3'''
		c = palette.color(role)
		return fvec3(c.red(), c.green(), c.blue()) / 255
	
	selection = mix(fvec3(0.4, 1, 0), qtc(palette.Highlight), 0.6)
	selection *= mix(1/max(selection), max(qtc(palette.Text)), 0.3)
	display.update({
		'background_color': qtc(palette.Base),
		'select_color_face': selection * 0.05,
		'select_color_line': selection * 1.1,
		})
	colors.update({
		'line': qtc(palette.Text),
		'point': qtc(palette.Text),
		'surface': mix(qtc(palette.Text), qtc(palette.Window), 0.7),
		'schematic': mix(qtc(palette.Text)*normalize(qtc(palette.LinkVisited)+0.01), qtc(palette.LinkVisited), 0.5),
		'annotation': mix(qtc(palette.Text)*normalize(qtc(palette.Link)+0.01), qtc(palette.Link), 0.5),
		})


# automatically load settings in the file exist
try:	load()
except FileNotFoundError:	pass



# author:	Jimy Byerley (jimy.byerley@gmail.com)
# license:	GNU-LGPLv3

""" Intends to replace the well known pprint, to format python objects on output
	
	nprint()      the pprint replacement: it works as print() but works on strings instead of working 
	              on already known objects split too long lines into indented blocks by syntax markers 
	              (such as {}()[], etc)
	              be careful:  as print(), nprint use the string conversion str(obj) to convert an object 
	                           passed, if you want to use the string representation, then you'll have to 
	                           pass repr(obj) to nprint or nformat
	nformat()     split a text into indented lines
	ncolor()      add color makers for terminals syntax highlighting
	deformat()   remove formatting of all marker-enclosed text
"""

from itertools import chain
import os, io


separators = ',;'
begin_marker = '{(['
end_marker = '})]'
symbols = ':*/+-<>'
specials = '\\'
spaces = ' \t\n'

line_width = 100

# blue-green colors
color_encloser =	'\x1b[38;5;42m'
color_marker =		'\x1b[38;5;46m'
color_specials =	'\x1b[38;5;33m'
color_number =		'\x1b[38;5;33m'
color_camel =		'\x1b[1;34;40m'

# red-orange colors
#color_encloser =	'\x1b[38;5;196m'
#color_marker =		'\x1b[38;5;214m'
#color_specials =	'\x1b[38;5;202m'
#color_number =		'\x1b[38;5;202m'
#color_camel =		'\x1b[1;34;40m'

# enable automatic coloration if supported
enable_color = 'COLORTERM' in os.environ


def _next_separator(text, seek):
	counter = 0
	i = seek
	for i in range(seek, len(text)):
		if text[i] in begin_marker:	counter += 1
		elif text[i] in end_marker:	counter -= 1
		if counter < 0 or (counter == 0 and text[i] in separators):	break
	return i

def _next_terminator(text, seek):
	counter = 0
	i = seek
	for i in range(seek, len(text)):
		if text[i] in begin_marker:	counter += 1
		elif text[i] in end_marker:	counter -= 1
		if counter < 0:		break
	return i
	
def _next_word(text, seek):
	while seek < len(text) and text[seek] in spaces:
		seek += 1
	return seek


def nformat(text, indent=0, width=None):
	''' output text, splited into indented blocks, newlines are kept '''
	if type(text) != str:	text = str(text)
	out = io.StringIO()
	seek = 0
	if not width:	width = line_width
	
	while seek < len(text):
		if text[seek] in begin_marker:
			term = _next_terminator(text, seek+1)
			if term - seek > width or '\n' in text[seek:term]:	# develop enclosement if its length is over that width
				indent += 1
				out.write(text[seek])
				if _next_separator(text, seek+1) < term: 	# no return if there is only one element in the enclosement
					out.write( '\n'+'\t'*indent )
					seek = _next_word(text, seek+1)-1
			else:
				out.write(text[seek:term+1])
				seek = term
		elif text[seek] in separators:
			out.write(text[seek] + '\n'+'\t'*indent)
			seek = _next_word(text, seek+1)-1
		elif text[seek] in end_marker:
			indent -= 1
			out.write(text[seek])
		else:
			out.write(text[seek])
		
		seek += 1
	
	return out.getvalue()
	
def deformat(text):
	''' remove formatting from text, leving only indentation and line breaks for lines not ended by a continuation character (,-+*/) '''
	seek = _next_word(text, 0)
	out = io.StringIO()
	
	while seek < len(text):
		if text[seek] in spaces:
			jump = _next_word(text, seek)
			if text[seek-1] in ',+-*/':
				out.write(' ')
			elif text[seek-1] in begin_marker or jump < len(text) and text[jump] in end_marker:
				pass
			elif '\n' in text[seek:jump]:
				out.write('\n')
			else:
				out.write(' ')
			seek = jump
		else:
			out.write(text[seek])
			seek += 1
	return out.getvalue()
		

def _bisect(l, index, key=lambda x:x):
	start,end = 0, len(l)
	while start < end:
		mid = (start+end)//2
		val = key(l[mid])
		if val < index:		start =	mid+1
		elif val > index:	end =	mid
		else:	return mid
	return start

def ncolor(text):
	''' generates a string with the content of the text passed, with terminal color makers to highlight some syntax elements '''
	# list of colors to apply
	coloration = []
	# look for special symbols
	for symbol in chain(separators, begin_marker, end_marker, symbols, specials):
		start = text.find(symbol, 0)
		while start != -1:
			end = start + len(symbol)
			color = color_encloser
			
			if symbol == '<':
				after_name = text.find(' ', end)
				if text[end:after_name].isidentifier():	
					end = after_name
				color = color_marker
			elif symbol == '>':			color = color_marker
			elif symbol in specials:	color = color_specials
			
			coloration.insert(
				_bisect(coloration, start, lambda x:x[1]), 
				(color, start, end-1)
				)
			start = text.find(symbol, end)
	
	# look for special worlds
	it = iter(text)
	i = 0
	for c in it:
		if c.isalnum():
			# look for numbers
			if c.isdigit():
				for j,c in enumerate(it):
					if not c.isalnum():	
						j -= 1
						break
				j += 1
				coloration.insert(
					_bisect(coloration, i, lambda x:x[1]), 
					(color_number, i, i+j)
					)
				i += j+1
			# look for CamelCase names
			#if c.isupper():
				#for j,c in enumerate(it):
					#if not c.isalnum():	
						#j -= 1
						#break
				#j += 1
				#coloration.insert(
					#_bisect(coloration, i, lambda x:x[1]), 
					#(color_camel, i, i+j)
					#)
				#i += j+1
		i += 1
	
	out = ''
	last = 0
	for color,start,end in coloration:
		end +=1
		out += text[last:start] + color + text[start:end] + '\x1b[0m'
		last = end
	out += text[last:]
	return out


def nprint(*args, indent=0, color=enable_color, end='\n'):
	""" write the arguments passed to the standard output, using nformat and ncolor """
	text = '\t'*indent + nformat(' '.join((str(arg) for arg in args)), indent)
	print(ncolor(text) if color else text, end=end)



if __name__ == '__main__':
	# for object dumping with string representation
	nprint(repr(dir()))
	# for common print use
	nprint('here is a list:', [[1000,2324,30],[40000,5000,6342342]], '\nhere is a type:', int, '\nhere is a Name in CamelCase and one in AVeryLongStringAndSoWhat')
	# for string and data output
	nprint('hello everyone, 100 is a decimal number and (0x27a) is an hexadecimal one between parentheses. (did you noticed the automatic line shift ?)')
	# numpy usage with repr() and without
	#import numpy as np
	#m = np.matrix([[1000,2324,30],[40000,5000,6342342]])
	#nprint(repr(m))
	#nprint(m)
	print(deformat('''
	P2 = vec3(0.08182,1.184e-08,-0.09931)
	P3 = vec3(-0.0431,1.258e-08,-0.1056)
P4 = vec3(-0.1593,-1.199e-08,0.1006)
line = [
	Segment(P4, P0),
	Segment(P0, P1 ),
	ArcThrough(P1, vec3(0.09681,-1.713e-09,0.01437), P2
		),
	Segment(P2,P3),
	ArcThrough(P3,vec3(-0.06933,-1.117e-09,0.009369),P4),
	]
axis = Axis(
	vec3(-0.1952,	4.918e-08,	-0.4126),
	vec3(1,	0,	0))
	'''))

	# hardcore test
	#import ast
	#nprint(ast.dump(ast.parse(open(__file__).read())))
